import operator
from typing import AbstractSet, Tuple

from . import expression_walker as ew
from . import type_system
from .exceptions import NeuroLangException
from .expression_pattern_matching import NeuroLangPatternMatchingNoMatch
from .expressions import (
    Constant,
    Definition,
    Expression,
    FunctionApplication,
    Symbol,
    Unknown, sure_is_not_pattern,
)
from .utils import NamedRelationalAlgebraFrozenSet, RelationalAlgebraSet
from .utils.relational_algebra_set import (
    RelationalAlgebraColumnInt, RelationalAlgebraColumnStr,
    RelationalAlgebraStringExpression
)

eq_ = Constant(operator.eq)


class Column:
    pass


class ColumnInt(int, Column):
    """Refer to a relational algebra set's column by its index."""


class ColumnStr(str, Column):
    """Refer to a named relational algebra set's column by its name."""


def str2columnstr_constant(name):
    return Constant[ColumnStr](
        ColumnStr(name), auto_infer_type=False, verify_type=False,
    )


def get_expression_columns(expression):
    columns = set()
    args = list(expression.unapply())
    while args:
        arg = args.pop()
        if isinstance(arg, Constant[Column]):
            columns.add(arg)
        elif isinstance(arg, Constant):
            continue
        elif isinstance(arg, Expression):
            args += arg.unapply()
        elif isinstance(arg, tuple):
            args += list(arg)

    return columns


class RelationalAlgebraOperation(Definition):
    def __init__(self):
        self._columns = set()

    def columns(self):
        if not hasattr(self, '_columns'):
            self._columns = get_expression_columns(self)
        return self._columns


class Selection(RelationalAlgebraOperation):
    def __init__(self, relation, formula):
        self.relation = relation
        self.formula = formula

    def __repr__(self):
        return f"\N{GREEK SMALL LETTER SIGMA}_{self.formula}({self.relation})"


class Projection(RelationalAlgebraOperation):
    def __init__(self, relation, attributes):
        self.relation = relation
        self.attributes = attributes

    def __repr__(self):
        return (
            f"\N{GREEK CAPITAL LETTER PI}"
            f"_{self.attributes}({self.relation})"
        )


class EquiJoin(RelationalAlgebraOperation):
    def __init__(
        self, relation_left, columns_left, relation_right, columns_right
    ):
        self.relation_left = relation_left
        self.columns_left = columns_left
        self.relation_right = relation_right
        self.columns_right = columns_right

    def __repr__(self):
        return (
            f"[{self.relation_left}"
            f"\N{JOIN}\N{SUBSCRIPT EQUALS SIGN}_{self.columns_left}"
            f"={self.columns_right}{self.relation_right}]"
        )


class NaturalJoin(RelationalAlgebraOperation):
    def __init__(self, relation_left, relation_right):
        self.relation_left = relation_left
        self.relation_right = relation_right

    def __repr__(self):
        return f"[{self.relation_left}" f"\N{JOIN}" f"{self.relation_right}]"


class Product(RelationalAlgebraOperation):
    def __init__(self, relations):
        self.relations = tuple(relations)

    def __repr__(self):
        return (
            "["
            + f"\N{n-ary times operator}".join(repr(r) for r in self.relations)
            + "]"
        )


class Difference(RelationalAlgebraOperation):
    def __init__(self, relation_left, relation_right):
        self.relation_left = relation_left
        self.relation_right = relation_right

    def __repr__(self):
        return f"[{self.relation_left}" f"-" f"{self.relation_right}]"


class Union(RelationalAlgebraOperation):
    def __init__(self, relation_left, relation_right):
        self.relation_left = relation_left
        self.relation_right = relation_right

    def __repr__(self):
        return f"{self.relation_left} ∪ {self.relation_right}"


class Intersection(RelationalAlgebraOperation):
    def __init__(self, relation_left, relation_right):
        self.relation_left = relation_left
        self.relation_right = relation_right

    def __repr__(self):
        return f"{self.relation_left} & {self.relation_right}"


class NameColumns(RelationalAlgebraOperation):
    """
    Give names to the columns of a relational algebra set.

    All columns must be named at once. Each column name must either be a
    `Constant[ColumnStr]` or a `Symbol[ColumnStr]` pointing to a symbolic
    column name resolved when the expression is compiled.

    """

    def __init__(self, relation, column_names):
        self.relation = relation
        self.column_names = column_names

    def __repr__(self):
        return (
            f"\N{GREEK SMALL LETTER DELTA}"
            f"_{self.column_names}({self.relation})"
        )


class RenameColumn(RelationalAlgebraOperation):
    def __init__(self, relation, src, dst):
        self.relation = relation
        self.src = src
        self.dst = dst

    def __repr__(self):
        return (
            f"\N{GREEK SMALL LETTER DELTA}"
            f"_({self.src}\N{RIGHTWARDS ARROW}{self.dst})"
            f"({self.relation})"
        )


class RenameColumns(RelationalAlgebraOperation):
    """
    Convenient operation for renaming multiple columns at the same time.

    Attributes
    ----------
    relation : NamedRelationalAlgebraFrozenSet
        The relation whose columns shall be renamed.
    renames : tuple of pairs of Constant[ColumnStr] or Symbol[ColumnStr]
        The renamings that should happen, represented as tuples (src, dst).

    """

    def __init__(self, relation, renames):
        self.relation = relation
        self.renames = renames

    def __repr__(self):
        return (
            f"\N{GREEK SMALL LETTER DELTA}"
            + "_({})".format(
                ", ".join(
                    "{}\N{RIGHTWARDS ARROW}{}".format(src, dst)
                    for src, dst in self.renames
                )
            )
            + f"({self.relation})"
        )


class ExtendedProjection(RelationalAlgebraOperation):
    """
    General operation defining string-based relational algebra projections
    allowing flexible computations on a relation's columns.

    Attributes
    ----------
    relation : Expression[AbstractSet]
        Relation on which the projections are applied.
    projection_list : Tuple[ExtendedProjectionListMember]
        List of projections to apply.

    Notes
    -----
    The concept of extended projection is formally defined in section 5.2.5
    of [1]_.

    .. [1] Garcia-Molina, Hector, Jeffrey D. Ullman, and Jennifer Widom.
       "Database systems: the complete book." (2009).

    """

    def __init__(self, relation, projection_list):
        self.relation = relation
        self.projection_list = tuple(projection_list)

    def __repr__(self):
        join_str = "," if len(self.projection_list) < 2 else ",\n"
        return "π_[{}]({})".format(
            join_str.join([repr(member) for member in self.projection_list]),
            repr(self.relation),
        )

    def __eq__(self, other):
        if not isinstance(other, ExtendedProjection):
            return False

        return (
            self.relation == other.relation and
            len(self.projection_list) == len(other.projection_list) and
            all(
                any(
                    element_self == element_other
                    for element_other in other.projection_list
                )
                for element_self in self.projection_list
            )
        )


class ExtendedProjectionListMember(Definition):
    """
    Member of a projection list.

    Attributes
    ----------
    fun_exp : `Constant[str]`
        Constant string representation of the extended projection operation.
    dst_column : `Constant[ColumnStr]` or `Symbol[ColumnStr]`
        Constant column string of the destination column.

    Notes
    -----
    As described in [1]_, a projection list member can either be
        - a single attribute (column) name in the relation, resulting in a
          normal non-extended projection,
        - an expression `x -> y` where `x` and `y` are both attribute (column)
          names, `x` effectively being rename as `y`,
        - or an expression `E -> z` where `E` is an expression involving
          attributes of the relation, arithmetic operators, and string
          operators, and `z` is a new name for the attribute that results from
          the calculation implied by `E`. For example, `a + b -> x` represents
          the sum of the attributes `a` and `b`, renamed `x`.

    .. [1] Garcia-Molina, Hector, Jeffrey D. Ullman, and Jennifer Widom.
       "Database systems: the complete book." (2009).

    """

    def __init__(self, fun_exp, dst_column):
        self.fun_exp = fun_exp
        self.dst_column = dst_column

    def __repr__(self):
        return "{} -> {}".format(self.fun_exp, self.dst_column)


class Destroy(RelationalAlgebraOperation):
    """
    Operation to map a column of a collection of elements into
    a new column with all collections concatenated

    Attributes
    ----------
    relation : Expression[AbstractSet]
        Relation on which the projections are applied.
    src_column : Constant[ColumnStr]
        Column to destroy in the operation.

    dst_column : Constant[ColumnStr]
        Column to map onto the new set.

    Notes
    -----
    The concept of set destruction is formally defined in chapter 20
    of [1]_.

    .. [1] Abiteboul, S., Hull, R. & Vianu, V. "Foundations of databases". (
           Addison Wesley, 1995).
    """

    def __init__(self, relation, src_column, dst_column):
        self.relation = relation
        self.src_column = src_column
        self.dst_column = dst_column

    def __repr__(self):
        return (
            f"set_destroy[{self.relation}, "
            "{self.src_column} -> {src.dst_column}]"
        )


class ConcatenateConstantColumn(RelationalAlgebraOperation):
    """
    Add a column with a repeated constant value to a relation.

    Attributes
    ----------
    relation : Constant[RelationalAlgebraSet]
        Relation to which the column will be added.

    column_name : Constant[ColumnStr] or Symbol[ColumnStr]
        Name of the newly added column.

    column_value : Constant
        Constant value repeated in the new column.

    """

    def __init__(self, relation, column_name, column_value):
        self.relation = relation
        self.column_name = column_name
        self.column_value = column_value


OPERATOR_STRING = {
    operator.add: "+",
    operator.sub: "-",
    operator.mul: "*",
    operator.truediv: "/",
    operator.eq: "==",
    operator.gt: ">",
    operator.lt: "<",
    operator.ge: ">=",
    operator.le: "<="
}


def is_arithmetic_operation(exp):
    """
    Whether the expression is an arithmetic operation function application.

    Parameters
    ----------
    exp : Expression

    Returns
    -------
    bool

    """
    return (
        isinstance(exp, FunctionApplication)
        and isinstance(exp.functor, Constant)
        and exp.functor.value
        in OPERATOR_STRING
    )


class StringArithmeticWalker(ew.PatternWalker):
    """
    Walker translating an Expression with basic arithmetic operations on a
    relation's columns to its equivalent string representation.

    The expression can refer to the names a relation's columns or to the
    length of an other constant relation.

    """
    @ew.add_match(FunctionApplication, is_arithmetic_operation)
    def arithmetic_operation(self, fa):
        return Constant[RelationalAlgebraStringExpression](
            RelationalAlgebraStringExpression(
                "({} {} {})".format(
                    self.walk(fa.args[0]).value,
                    OPERATOR_STRING[fa.functor.value],
                    self.walk(fa.args[1]).value,
                ),
            ),
            auto_infer_type=False,
            verify_type=False,
        )

    @ew.add_match(Constant[ColumnStr])
    def constant_column_str(self, cst_col_str):
        return Constant[RelationalAlgebraStringExpression](
            RelationalAlgebraStringExpression(cst_col_str.value),
            auto_infer_type=False,
            verify_type=False,
        )

    @ew.add_match(Constant[int])
    def constant_int(self, cst):
        return Constant[RelationalAlgebraStringExpression](
            str(cst.value),
            auto_infer_type=False,
            verify_type=False,
        )

    @ew.add_match(Constant[float])
    def constant_float(self, cst):
        return Constant[RelationalAlgebraStringExpression](
            str(cst.value),
            auto_infer_type=False,
            verify_type=False,
        )

    @ew.add_match(Constant[str])
    def constant_str(self, cst):
        return Constant[RelationalAlgebraStringExpression](
            f'"{cst.value}"',
            auto_infer_type=False,
            verify_type=False,
        )


class ReplaceConstantColumnStrBySymbol(ew.ExpressionWalker):
    @ew.add_match(Constant[ColumnStr])
    def column_str(self, expression):
        return Symbol[ColumnStr](expression.value)


class RelationalAlgebraSolver(ew.ExpressionWalker):
    """
    Mixing that walks through relational algebra expressions and
    executes the operations.

    Relations are expected to be represented
    as objects with the same interface as :obj:`RelationalAlgebraSet`.
    """

    _rccsbs = ReplaceConstantColumnStrBySymbol()
    _saw = StringArithmeticWalker()
    _fa_2_lambda = ew.FunctionApplicationToPythonLambda()

    def __init__(self, symbol_table=None):
        self.symbol_table = symbol_table

    @ew.add_match(
        Selection(
            Constant, FunctionApplication(
                eq_,
                (Constant[Column], Constant[Column])
            )
        )
    )
    def selection_between_columns(self, selection):
        col1, col2 = selection.formula.args
        selected_relation = (
            selection.relation
        ).value.selection_columns({col1.value: col2.value})

        return self._build_relation_constant(selected_relation)

    def _build_relation_constant(self, relation, type_=Unknown):
        if type_ is not Unknown:
            relation_type = type_
        else:
            relation_type = _infer_relation_type(relation)
        return Constant[relation_type](relation, verify_type=False)

    @ew.add_match(
        Selection(Constant, FunctionApplication(eq_, (Constant[Column], ...)))
    )
    def selection_by_constant(self, selection):
        col, val = selection.formula.args
        selected_relation = selection.relation.value.selection(
            {col.value: val.value}
        )

        return self._build_relation_constant(selected_relation)

    @ew.add_match(
        Selection(Constant, FunctionApplication)
    )
    def selection_general_selection_by_constant(self, selection):
        relation = selection.relation
        compiled_formula = self._compile_function_application_to_sql_fun_exp(
            selection.formula
        )
        selected_relation = relation.value.selection(compiled_formula)
        return self._build_relation_constant(selected_relation)

    @ew.add_match(Projection(Constant, ...))
    def ra_projection(self, projection):
        relation = projection.relation
        cols = tuple(v.value for v in projection.attributes)
        projected_relation = relation.value.projection(*cols)
        return self._build_relation_constant(projected_relation)

    @ew.add_match(
        Product,
        lambda product: all(
            isinstance(relation, Constant)
            for relation in product.relations
        )
    )
    def ra_product(self, product):
        if len(product.relations) == 0:
            return Constant[AbstractSet](RelationalAlgebraSet(set()))

        res = self.walk(product.relations[0]).value
        for relation in product.relations[1:]:
            res = res.cross_product(self.walk(relation).value)
        return self._build_relation_constant(res)

    @ew.add_match(EquiJoin(Constant, ..., Constant, ...))
    def ra_equijoin(self, equijoin):
        left = equijoin.relation_left.value
        columns_left = (c.value for c in equijoin.columns_left)
        right = equijoin.relation_right.value
        columns_right = (c.value for c in equijoin.columns_right)
        res = left.equijoin(right, list(zip(columns_left, columns_right)))

        return self._build_relation_constant(res)

    @ew.add_match(NaturalJoin(Constant, Constant))
    def ra_naturaljoin(self, naturaljoin):
        left = naturaljoin.relation_left.value
        right = naturaljoin.relation_right.value
        res = left.naturaljoin(right)
        return self._build_relation_constant(res)

    @ew.add_match(Difference(Constant, Constant))
    def ra_difference(self, difference):
        return self._type_preserving_binary_operation(difference)

    @ew.add_match(Union(Constant, Constant))
    def ra_union(self, union):
        return self._type_preserving_binary_operation(union)

    @ew.add_match(Intersection(Constant, Constant))
    def ra_intersection(self, intersection):
        return self._type_preserving_binary_operation(intersection)

    @ew.add_match(NameColumns(Constant, ...))
    def ra_name_columns(self, name_columns):
        relation = name_columns.relation
        relation_set = relation.value
        column_names = tuple(
            self.walk(column_name).value
            for column_name in name_columns.column_names
        )
        new_set = NamedRelationalAlgebraFrozenSet(column_names, relation_set)
        return self._build_relation_constant(new_set)

    @ew.add_match(RenameColumn(Constant, ..., ...))
    def ra_rename_column(self, rename_column):
        relation = rename_column.relation
        src = rename_column.src.value
        dst = rename_column.dst.value
        new_set = relation.value
        new_set = new_set.rename_column(src, dst)
        return self._build_relation_constant(new_set)

    @ew.add_match(RenameColumns(Constant, ...))
    def ra_rename_columns(self, rename_columns):
        if len(set(c for c, _ in rename_columns.renames)) < len(
            rename_columns.renames
        ):
            raise ValueError("Cannot have duplicated source columns")
        relation = rename_columns.relation
        new_set = relation.value
        renames = {
            src.value: dst.value for src, dst in rename_columns.renames
        }
        new_set = new_set.rename_columns(renames)
        return self._build_relation_constant(new_set)

    @ew.add_match(ConcatenateConstantColumn(Constant, ..., ...))
    def concatenate_constant_column(self, concat_op):
        relation = concat_op.relation
        new_column = concat_op.column_name
        new_column_value = concat_op.column_value
        if new_column.value in relation.value.columns:
            raise NeuroLangException(
                "Cannot concatenate column to a relation that already "
                "has a column with that name. Same column name: {}".format(
                    new_column
                )
            )
        ext_proj_list_members = []
        for column in relation.value.columns:
            cst_column = Constant[ColumnStr](
                column, auto_infer_type=False, verify_type=False
            )
            ext_proj_list_members.append(
                ExtendedProjectionListMember(
                    fun_exp=cst_column, dst_column=cst_column,
                )
            )
        ext_proj_list_members.append(
            ExtendedProjectionListMember(
                fun_exp=new_column_value, dst_column=new_column
            )
        )
        return self.walk(ExtendedProjection(relation, ext_proj_list_members))

    @ew.add_match(ExtendedProjection(Constant, ...))
    def extended_projection(self, proj_op):
        relation = proj_op.relation
        eval_expressions = {}
        for member in proj_op.projection_list:
            fun_exp = self.walk(member.fun_exp)
            eval_expressions[
                member.dst_column.value
            ] = self._compile_function_application_to_sql_fun_exp(fun_exp)
        with sure_is_not_pattern():
            result = relation.value.extended_projection(eval_expressions)
        return self._build_relation_constant(result)

    def _compile_function_application_to_sql_fun_exp(self, fun_exp):
        if isinstance(fun_exp, FunctionApplication):
            try:
                return self._saw.walk(fun_exp).value
            except NeuroLangPatternMatchingNoMatch:
                fun, args = self._fa_2_lambda.walk(self._rccsbs.walk(fun_exp))
                return lambda t: fun(
                    **{arg: t[arg] for arg in args}
                )
        elif isinstance(fun_exp, Constant[ColumnInt]):
            return RelationalAlgebraColumnInt(fun_exp.value)
        elif isinstance(fun_exp, Constant[ColumnStr]):
            return RelationalAlgebraColumnStr(fun_exp.value)
        else:
            return fun_exp.value

    @ew.add_match(FunctionApplication, is_arithmetic_operation)
    def prov_arithmetic_operation(self, arithmetic_op):
        args = self.walk(arithmetic_op.args)
        if any(
            arg_new is not arg_old
            for arg_new, arg_old in zip(args, arithmetic_op.args)
        ):
            return FunctionApplication[arithmetic_op.type](
                arithmetic_op.functor, args
            )
        else:
            return arithmetic_op

    @ew.add_match(Destroy(Constant, ..., ...))
    def set_destroy(self, destroy):
        relation = destroy.relation.value
        src_column = self.walk(destroy.src_column).value
        dst_columns = self.walk(destroy.dst_column).value
        if src_column not in relation.columns:
            raise NeuroLangException(
                f"source column {src_column} not present in "
                "set's columns"
            )

        cols = relation.columns
        set_type = type(relation)
        if not isinstance(dst_columns, tuple):
            dst_columns = (dst_columns,)
        dst_cols = cols + tuple(d for d in dst_columns if d not in cols)
        result_set = set_type(columns=dst_cols)
        if len(cols) > 0:
            row_group_iterator = (t for _, t in relation.groupby(cols))
        else:
            row_group_iterator = (relation,)
        for t in row_group_iterator:
            destroyed_set = set_type(columns=dst_columns)
            for row in t:
                row_set = set_type(
                    columns=dst_columns,
                    iterable=getattr(row, src_column)
                )
                destroyed_set = destroyed_set | row_set
            new_set = (
                t
                .projection(*cols)
                .naturaljoin(destroyed_set)
            )
            result_set = result_set | new_set
        return self._build_relation_constant(result_set)

    @ew.add_match(Constant)
    def ra_constant(self, constant):
        return constant

    @ew.add_match(Symbol)
    def ra_symbol(self, symbol):
        try:
            constant = self.symbol_table[symbol]
        except KeyError:
            raise NeuroLangException(f"Symbol {symbol} not in table")
        return constant

    @ew.add_match(Constant[RelationalAlgebraStringExpression])
    def arithmetic_string_expression(self, expression):
        return expression

    def _type_preserving_binary_operation(self, ra_op):
        """
        Generic function to apply binary operations (A <op> B) where A and B's
        tuples have the same type, and whose results's tuples have the same
        type as A and B's tuples. This includes, but is not necessarily limited
        to, the Union, Intersection and Difference operations.
        """
        left = self.walk(ra_op.relation_left)
        right = self.walk(ra_op.relation_right)
        left_type = _get_const_relation_type(left)
        right_type = _get_const_relation_type(right)
        type_ = type_system.unify_types(left_type, right_type)
        binary_op_fun_name = {
            Union: "__or__",
            Intersection: "__and__",
            Difference: "__sub__",
        }.get(type(ra_op))
        new_relation = getattr(left.value, binary_op_fun_name)(right.value)
        return self._build_relation_constant(new_relation, type_=type_)


class RelationalAlgebraSimplification(ew.ExpressionWalker):
    @ew.add_match(Product, lambda x: len(x.relations) == 1)
    def single_product(self, product):
        return self.walk(product.relations[0])


class RelationalAlgebraRewriteSelections(ew.ExpressionWalker):
    """
    Mixing that optimises through relational algebra expressions.

    The pushing selections (:obj:`Selection`) down and compositions
    and reduces the reach of selections. Then it converts
    equi-selection/product compositions into equijoins.
    """

    @ew.add_match(
        Selection(..., FunctionApplication(eq_, (..., Constant[Column]))),
        lambda s: not issubclass(s.formula.args[0].type, Column),
    )
    def swap_formula_args(self, selection):
        new_selection = Selection(
            self.walk(selection.relation), eq_(*selection.formula.args[::-1])
        )
        return self.walk(new_selection)

    @ew.add_match(
        Selection(
            ..., FunctionApplication(eq_, (Constant[Column], Constant[Column]))
        ),
        lambda s: s.formula.args[0].value > s.formula.args[1].value,
    )
    def sort_formula_args(self, selection):
        new_selection = Selection(
            self.walk(selection.relation), eq_(*selection.formula.args[::-1])
        )
        return self.walk(new_selection)

    @ew.add_match(
        Selection(
            Selection(..., FunctionApplication(eq_, ...)),
            FunctionApplication(eq_, ...),
        ),
        lambda s: s.formula.args[0].value > s.relation.formula.args[0].value,
    )
    def selection_selection_swap(self, selection):
        new_selection = Selection(
            Selection(selection.relation.relation, selection.formula),
            selection.relation.formula,
        )
        return self.walk(new_selection)

    @ew.add_match(
        Selection(
            Product, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        ),
        lambda s: (
            s.formula.args[0].value
            >= RelationalAlgebraRewriteSelections.get_arity(
                s.relation.relations[0]
            )
        ),
    )
    def selection_push_right(self, selection):
        relations = selection.relation.relations
        column = int(selection.formula.args[0].value)

        i, accum_arity, column = self.split_relations_column(relations, column)

        left_relations = relations[:i]
        relations = relations[i:]

        arg_right = selection.formula.args[1]
        if issubclass(arg_right.type, Column):
            arg_right = Constant[ColumnInt](
                ColumnInt(-accum_arity + arg_right.value)
            )

        res = Product(
            left_relations
            + (
                Selection(
                    Product(relations),
                    eq_(Constant[ColumnInt](ColumnInt(column)), arg_right),
                ),
            )
        )
        return self.walk(res)

    @ew.add_match(
        Selection(
            Product, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        ),
        lambda s: (
            s.formula.args[1].value
            < (
                RelationalAlgebraRewriteSelections.get_arity(s.relation)
                - RelationalAlgebraRewriteSelections.get_arity(
                    s.relation.relations[-1]
                )
            )
        ),
    )
    def selection_shorten_right(self, selection):
        relations = selection.relation.relations
        column = int(selection.formula.args[1].value)

        i, accum_arity, column = self.split_relations_column(relations, column)
        column += accum_arity
        inner_relations = tuple(relations[: i + 1])
        if len(inner_relations) == 1:
            inner_relations = inner_relations[0]
        else:
            inner_relations = Product(inner_relations)

        outer_relations = tuple(relations[i + 1:])

        arg_left = selection.formula.args[0]

        res = Product(
            (
                Selection(
                    inner_relations,
                    eq_(arg_left, Constant[ColumnInt](ColumnInt(column))),
                ),
            )
            + outer_relations
        )
        return self.walk(res)

    @ew.add_match(
        Selection(
            EquiJoin,
            FunctionApplication(
                eq_, (Constant[ColumnInt], Constant[ColumnInt])
            ),
        )
    )
    def selection_on_equijoin_columns(self, selection):
        column_left = selection.formula.args[0].value
        column_right = selection.formula.args[1].value
        column_min, column_max = sorted((column_left, column_right))
        relation_left = selection.relation.relation_left
        relation_right = selection.relation.relation_right
        left_arity = self.get_arity(relation_left)
        if column_max < left_arity:
            relation_left = Selection(relation_left, selection.formula)
        elif column_min >= left_arity:
            column_min -= left_arity
            column_max -= left_arity
            new_formula = eq_(
                Constant[ColumnInt](ColumnInt(column_min)),
                Constant[ColumnInt](ColumnInt(column_max)),
            )
            relation_right = Selection(relation_right, new_formula)
        else:
            return selection

        return self.walk(
            EquiJoin(
                relation_left,
                selection.relation.columns_left,
                relation_right,
                selection.relation.columns_right,
            )
        )

    @ew.add_match(
        Selection(
            EquiJoin, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        )
    )
    def selection_on_equijoin(self, selection):
        column = selection.formula.args[0].value
        relation_left = selection.relation.relation_left
        relation_right = selection.relation.relation_right
        left_arity = self.get_arity(relation_left)
        if column < left_arity:
            relation_left = Selection(relation_left, selection.formula)
        else:
            relation_right = Selection(
                relation_right,
                eq_(
                    Constant(
                        ColumnInt(
                            int(selection.formula.args[0].value) - left_arity
                        )
                    ),
                    selection.formula.args[1],
                ),
            )

        return self.walk(
            EquiJoin(
                relation_left,
                selection.relation.columns_left,
                relation_right,
                selection.relation.columns_right,
            )
        )

    @ew.add_match(
        Selection(
            Product,
            FunctionApplication(
                eq_, (Constant[ColumnInt], Constant[ColumnInt])
            ),
        )
    )
    def selection_between_columns_product(self, selection):
        relations = selection.relation.relations
        if len(relations) == 1:
            res = Selection(relations[0], selection.formula)
        else:
            column_left = (selection.formula.args[0],)

            column_right = int(selection.formula.args[1].value)
            left_arity = self.get_arity(relations[0])
            column_right -= left_arity
            column_right = (Constant[ColumnInt](ColumnInt(column_right)),)

            relations_right = relations[1:]
            if len(relations_right) == 1:
                relation_right = relations_right[0]
            else:
                relation_right = Product(relations_right)

            res = EquiJoin(
                relations[0], column_left, relation_right, column_right
            )

        return self.walk(res)

    @ew.add_match(
        Selection(
            Product, FunctionApplication(eq_, (Constant[ColumnInt], ...))
        )
    )
    def selection_by_constant_on_product(self, selection):
        return self.walk(
            Product(
                (
                    Selection(
                        selection.relation.relations[0], selection.formula
                    ),
                )
                + selection.relation.relations[1:]
            )
        )

    @staticmethod
    def split_relations_column(relations, column):
        accum_arity = 0
        for i, relation in enumerate(relations):
            current_arity = RelationalAlgebraRewriteSelections.get_arity(
                relation
            )
            if column < current_arity:
                break
            accum_arity += current_arity
            column -= current_arity
        return i, accum_arity, column

    @staticmethod
    def get_arity(expression):
        if isinstance(expression, Constant):
            return expression.value.arity
        elif isinstance(expression, Product):
            return sum(
                RelationalAlgebraRewriteSelections.get_arity(r)
                for r in expression.relations
            )
        elif isinstance(expression, EquiJoin):
            return RelationalAlgebraRewriteSelections.get_arity(
                expression.relation_left
            ) + RelationalAlgebraRewriteSelections.get_arity(
                expression.relation_right
            )
        else:
            return RelationalAlgebraRewriteSelections.get_arity(
                expression.relation
            )


class RelationalAlgebraOptimiser(
    RelationalAlgebraRewriteSelections,
    RelationalAlgebraSimplification,
    ew.ExpressionWalker,
):
    """
    Mixing that optimises through relational algebra expressions by
    rewriting.
    equi-selection/product compositions into equijoins.
    """

    pass


def _const_relation_type_is_known(const_relation):
    """
    Returns whether `T` in `Constant[T]` matches `AbstractSet[Tuple[type_1,
    ..., type_n]]`, in which case we consider the type of the relation's tuples
    to be known.

    """
    type_args = type_system.get_args(const_relation.type)
    if len(type_args) != 1:
        return False
    tuple_type_args = type_system.get_args(type_args[0])
    return len(tuple_type_args) > 0


def _sort_typed_const_named_relation_tuple_type_args(const_named_relation):
    """
    Given a typed `Constant[NamedRelationalAlgebraFrozenSet]`, `R`, with
    columns `c_1, ..., c_n` and whose tuples have the type `Tuple[x_1, ...,
    x_n]`, this function obtains the new type of the relation
    `AbstractSet[Tuple[y_1, ..., y_n]]` such that `y_1, ..., y_n` are the same
    initial types `x_1, ..., x_n` but sorted based on the alphabetical sort of
    columns `c_1, ..., c_n`.

    Notes
    -----
    This is useful when comparing or unifying the types of two named relations
    that have a different column sorting.

    """
    tuple_args = const_named_relation.type.__args__[0].__args__
    columns = const_named_relation.value.columns
    sorted_tuple_args = tuple(
        tuple_args[i]
        for i, _ in sorted(enumerate(columns), key=lambda x: x[1])
    )
    return AbstractSet[Tuple[sorted_tuple_args]]


def _infer_relation_type(relation):
    """
    Infer the type of the tuples in the relation based on its first tuple. If
    the relation is empty, just return `Abstract[Tuple]`.
    """
    if relation.is_empty() or relation.arity == 0:
        return AbstractSet[Tuple]
    if hasattr(relation, "row_type"):
        return AbstractSet[relation.row_type]
    tuple_type = Tuple[tuple(type(arg) for arg in relation.fetch_one())]
    return AbstractSet[tuple_type]


def _get_const_relation_type(const_relation):
    if _const_relation_type_is_known(const_relation):
        if isinstance(const_relation.value, NamedRelationalAlgebraFrozenSet):
            return _sort_typed_const_named_relation_tuple_type_args(
                const_relation
            )
        else:
            return const_relation.type
    else:
        return _infer_relation_type(const_relation.value)


class EliminateTrivialProjections(ew.PatternWalker):
    @ew.add_match(Projection(Constant, ...))
    def eliminate_trivial_projection(self, expression):
        if (
            tuple(c.value for c in expression.attributes) ==
            tuple(c for c in expression.relation.value.columns)
        ):
            return expression.relation
        else:
            return expression


class RelationalAlgebraPushInSelections(ew.PatternWalker):
    @ew.add_match(
        Selection(NaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_right)
            ) == 0
        )
    )
    def push_selection_in_left(self, expression):
        return self.walk(
            NaturalJoin(
                Selection(
                    expression.relation.relation_left,
                    expression.formula
                ),
                expression.relation.relation_right
            )
        )

    @ew.add_match(
        Selection(NaturalJoin, ...),
        lambda exp: (
            len(
                get_expression_columns(exp.formula) &
                get_expression_columns(exp.relation.relation_left)
            ) == 0
        )
    )
    def push_selection_in_right(self, expression):
        return self.walk(
            NaturalJoin(
                expression.relation.relation_left,
                Selection(
                    expression.relation.relation_right,
                    expression.formula
                )
            )
        )

    @ew.add_match(
        Selection(Projection, ...),
        lambda exp: len(
            set(exp.relation.attributes) &
            get_expression_columns(exp.formula)
        ) == 0
    )
    def push_selection_in_projection(self, expression):
        return Projection(
            Selection(expression.relation.relation, expression.formula),
            expression.relation.attributes
        )



from ...expressions import (
    Symbol,
    Constant,
    ExpressionBlock,
    FunctionApplication,
)
from ...datalog.expressions import Fact
from ...logic import (
    Implication,
    Conjunction,
    UniversalPredicate,
)
from ...logic.horn_clauses import fol_query_to_datalog_program
from .drs_builder import DRSBuilder, DRS2FOL
from .chart_parser import ChartParser
from .english_grammar import EnglishGrammar, EnglishBaseLexicon
from .exceptions import TranslateToDatalogException
import operator
from ...expression_walker import ExpressionWalker
from ...logic.transformations import (
    CollapseConjunctions,
    DistributeUniversalQuantifiers,
    DistributeImplicationsWithConjunctiveHeads,
)


_equals = Constant(operator.eq)


class TranslateToDatalog:
    def __init__(self):
        self.parser = ChartParser(EnglishGrammar, EnglishBaseLexicon())
        self.builder = DRSBuilder(EnglishGrammar)
        self.into_fol = DRS2FOL()

    def translate_block(self, string):
        program = ExpressionBlock(())

        for sentence in string.split("."):
            sentence = sentence.strip()
            if not sentence:
                continue
            program = ExpressionBlock(
                program.expressions
                + self.translate_sentence(sentence).expressions
            )

        return program

    def translate_sentence(self, sentence):
        t = self.parser.parse(sentence)[0]

        drs = self.builder.walk(t)
        exp = self.into_fol.walk(drs)
        exp = TransformIntoConjunctionOfDatalogSentences().walk(exp)

        lsentences = exp.formulas if isinstance(exp, Conjunction) else (exp,)
        program = ExpressionBlock(())
        for block in map(self.translate_logical_sentence, lsentences):
            program = ExpressionBlock(program.expressions + block.expressions)
        return program

    def translate_logical_sentence(self, exp):
        try:
            return _as_intensional_rule(exp)
        except TranslateToDatalogException:
            pass

        try:
            return _as_fact(exp)
        except TranslateToDatalogException:
            pass

        raise TranslateToDatalogException(
            f"Unsupported expression: {repr(exp)}"
        )


class TransformIntoConjunctionOfDatalogSentences(
    DistributeImplicationsWithConjunctiveHeads,
    DistributeUniversalQuantifiers,
    CollapseConjunctions,
    ExpressionWalker,
):
    """
    A datalog-sentence in this case is a logical sentence which can be
    interpreted as datalog. The only 2 types of sentences supported are facts
    and rules. This rewrite allows to use conjunctions in a more flexible way,
    allowing to use them between facts and in implication heads, because then
    they will be properly distributed.
    """

    pass


def _as_intensional_rule(exp):
    ucv, exp = _strip_universal_quantifiers(exp)

    if not isinstance(exp, Implication):
        raise TranslateToDatalogException(
            "A Datalog rule must be an implication"
        )

    head = exp.consequent
    body = exp.antecedent

    if not isinstance(head, FunctionApplication):
        raise TranslateToDatalogException(
            "The head of a Datalog rule must be a function application"
        )

    head, body, ucv = _constrain_using_head_constants(head, body, ucv)

    if any(a not in ucv for a in head.args):
        raise TranslateToDatalogException(
            "All rule head arguments must be universally quantified"
        )

    return fol_query_to_datalog_program(head, body)


def _strip_universal_quantifiers(exp):
    ucv = ()

    while isinstance(exp, UniversalPredicate):
        ucv += (exp.head,)
        exp = exp.body

    return ucv, exp


def _add_universal_quantifiers(exp, ucv):
    ucv = list(ucv)
    while ucv:
        v = ucv.pop()
        exp = UniversalPredicate(v, exp)
    return exp


def _constrain_using_head_constants(head, body, ucv):
    args = ()

    for a in head.args:
        if isinstance(a, Constant):
            s = Symbol.fresh()
            body = Conjunction((body, _equals(s, a)))
            ucv += (s,)
            args += (s,)
        else:
            args += (a,)

    head = head.functor(*args)
    return head, body, ucv


def _as_fact(exp):
    if not isinstance(exp, FunctionApplication):
        raise TranslateToDatalogException(
            "A fact must be a single function application"
        )

    return ExpressionBlock((Fact(exp),))



from .. import expressions as exp
from . import UnaryLogicOperator, NaryLogicOperator


def most_general_unifier(expression1, expression2):
    '''
    Obtain the most general unifier (MGU) between two function applications.
    If the MGU exists it returns the substitution and the unified expression.
    If the MGU doesn't exist it returns None.
    '''
    args1, args2 = most_general_unifier_extract_arguments(
        expression1, expression2
    )

    if args1 is None or args2 is None:
        return None

    unifier = most_general_unifier_arguments(args1, args2)

    if unifier is None:
        return unifier
    else:
        return (
            unifier[0],
            apply_substitution(expression1, unifier[0])
        )


def most_general_unifier_extract_arguments(expression1, expression2):
    expression_stack = [(expression1, expression2)]
    args1 = tuple()
    args2 = tuple()
    while expression_stack:
        expression1, expression2 = expression_stack.pop(0)
        if not (
            expression1.functor == expression2.functor and
            len(expression1.args) == len(expression2.args)
        ):
            return None, None

        for arg1, arg2 in zip(expression1.args, expression2.args):
            is_application1 = isinstance(arg1, exp.FunctionApplication)
            is_application2 = isinstance(arg2, exp.FunctionApplication)
            if is_application1 and is_application2:
                expression_stack.append((arg1, arg2))
            elif is_application1 or is_application2:
                return None, None
            else:
                args1 += (arg1,)
                args2 += (arg2,)

    return args1, args2


def apply_substitution(function_application, substitution):
    if isinstance(function_application, UnaryLogicOperator):
        return function_application.apply(*(
            apply_substitution(formula, substitution)
            for formula in function_application.unapply()
        ))
    if isinstance(function_application, NaryLogicOperator):
        return function_application.apply(
            *(
                apply_substitution_arguments(formula, substitution)
                for formula in function_application.unapply()
            )
        )
    return type(function_application)(
        function_application.functor,
        apply_substitution_arguments(function_application.args, substitution)
    )


def most_general_unifier_arguments(args1, args2):
    '''
    Obtain the most general unifier (MGU) between argument tuples.
    If the MGU exists it returns the substitution and the unified arguments.
    If the MGU doesn't exist it returns None.
    '''
    if not (
        isinstance(args1, tuple) and
        isinstance(args2, tuple)
    ):
        return ValueError("We can only unify argument tuples")

    if len(args1) != len(args2):
        return None

    substitution = dict()
    while True:
        for arg1, arg2 in zip(args1, args2):
            if arg1 != arg2:
                break
        else:
            return substitution, args1

        if isinstance(arg1, exp.Symbol):
            substitution[arg1] = arg2
        elif isinstance(arg2, exp.Symbol):
            substitution[arg2] = arg1
        else:
            return None

        args1 = apply_substitution_arguments(args1, substitution)
        args2 = apply_substitution_arguments(args2, substitution)


def apply_substitution_arguments(arguments, substitution):
    new_args = tuple()
    for a in arguments:
        if isinstance(a, exp.FunctionApplication):
            new_args += (apply_substitution(a, substitution),)
        else:
            new_args += (substitution.get(a, a),)
    return new_args


def merge_substitutions(subs1, subs2):
    if len(subs1) > len(subs2):
        aux = subs1
        subs1 = subs2
        subs2 = aux

    if any(
        v != subs2[k]
        for k, v in subs1.items()
        if k in subs2
    ):
        return None
    res = subs1.copy()
    res.update(subs2)
    return res


def compose_substitutions(subs1, subs2):
    new_subs = dict()
    new_subs = {
        k: v for k, v in subs2.items()
        if k not in subs1
    }

    for k, v in subs1.items():
        if isinstance(v, exp.Symbol):
            new_value = subs2.get(k, v)
            if new_value != k:
                new_subs[k] = new_value
        else:
            new_subs[k] = v

    return new_subs



from pytest import raises

from .. import expression_walker
from .. import expressions

from typing import Callable, AbstractSet, Tuple

S_ = expressions.Symbol
C_ = expressions.Constant
F_ = expressions.FunctionApplication


def test_symbol_table_as_parameter():
    symbol_table = expressions.TypedSymbolTable()
    solver = expression_walker.TypedSymbolTableEvaluator(symbol_table)
    s = S_('S1')
    c = C_[str]('S')
    symbol_table[s] = c
    assert solver.symbol_table[s] is c


def test_expression_evaluator():
    ebe = expression_walker.ExpressionBasicEvaluator()
    a = C_(1)
    b = C_(2)

    c = ebe.walk(a + b)
    assert isinstance(c, C_[int])
    assert c.value == 3


def test_evaluating_embedded_functions():
    class Walker(
        expression_walker.ExpressionBasicEvaluator,
        expression_walker.TypedSymbolTableEvaluator
    ):
        def function_add_set(self, a: AbstractSet[int]) -> int:
            return sum(a)

        def function_add_tuple(self, a: Tuple[int, int]) -> int:
            return sum(a)

    a = C_[AbstractSet[int]](frozenset(
        C_[int](i) for i in range(3)
    ))

    w = Walker()
    f = S_[Callable[[AbstractSet[int]], int]]('add_set')

    assert f in w.symbol_table
    assert isinstance(w.symbol_table[f], expressions.Constant)
    assert w.symbol_table[f].value == w.function_add_set

    res = w.walk(F_[int](f, (a,)))
    assert res.value == 3

    f = S_[Callable[[AbstractSet[int]], int]]('add_tuple')
    a = C_[Tuple[int, int, int]](tuple(
        C_[int](i) for i in range(3)
    ))

    res = w.walk(F_[int](f, (a,)))
    assert res.value == 3


def test_pattern_walker_wrong_args():
    with raises(TypeError):
        class PM(expression_walker.PatternWalker):
            @expression_walker.add_match(
                F_(...)
            )
            def __(self, expression):
                return expression


def test_symbol_table_scopes():
    symbol_table = expressions.TypedSymbolTable()
    solver = expression_walker.TypedSymbolTableEvaluator(symbol_table)
    s = S_('S1')
    c = C_[str]('S')
    symbol_table[s] = c
    assert solver.symbol_table[s] is c
    solver.push_scope()
    assert len(solver.symbol_table) == 0
    assert solver.symbol_table[s] == c
    solver.pop_scope()
    assert solver.symbol_table[s] is c
    with raises(expressions.NeuroLangException):
        solver.pop_scope()
        solver.pop_scope()


def test_entry_point_walker():
    class PM(expression_walker.EntryPointPatternWalker):
        @expression_walker.add_entry_point_match(F_(S_('start'), ...))
        def entry_point(self, expression):
            return expression.functor(*(
                self.walk(arg) for arg in expression.args
            ))

        @expression_walker.add_match(F_)
        def __(self, expression):
            return expression

    pm = PM()

    exp_correct = S_('start')(S_('a')())
    exp_wrong = S_('end')(S_('a')())

    res = pm.walk(exp_correct)
    assert res == exp_correct

    with raises(expressions.NeuroLangException):
        pm.walk(exp_wrong)

    class PM2(PM):
        @expression_walker.add_entry_point_match(
            F_,
            lambda x: x.functor == S_('pre_start')
        )
        def entry_point(self, expression):
            return expression.functor(*(
                self.walk(arg) for arg in expression.args
            ))

    pm2 = PM2()

    exp_pre_correct = S_('pre_start')(S_('a')())
    res = pm2.walk(exp_pre_correct)

    assert res == exp_pre_correct

    with raises(expressions.NeuroLangException):
        pm2.walk(exp_correct)

    with raises(
        expressions.NeuroLangException, match="Entry point not declared"
    ):

        class PM3(expression_walker.EntryPointPatternWalker):
            @expression_walker.add_match(F_)
            def __(self, expression):
                return expression

        PM3()


def test_chained_walker():
    class WalkerA(expression_walker.ExpressionWalker):
        @expression_walker.add_match(S_("A"))
        def match_a(self, expression):
            return S_("B")

    class WalkerB(expression_walker.ExpressionWalker):
        @expression_walker.add_match(S_("A"))
        def match_a(self, expression):
            return S_("Q")

        @expression_walker.add_match(S_("B"))
        def match_a(self, expression):
            return S_("C")

    walker = expression_walker.ChainedWalker(WalkerA, WalkerB)
    exp = S_("A")
    res = walker.walk(exp)
    assert res == S_("C")


def test_convert_to_lambda():
    add = C_(lambda x, y: x + y)
    func = add(S_('x'), C_(1))

    fa2pl = expression_walker.FunctionApplicationToPythonLambda()

    l0, args = fa2pl.walk(func)
    assert l0(x=3) == 4
    assert args == {'x'}

    func = add(S_('x'), S_('y'))
    l1, args = fa2pl.walk(func)
    assert l1(x=3, y=1) == 4
    assert args == {'x', 'y'}

    sub = C_(lambda x, y: x - y)
    func = sub(add(S_('x'), C_(1)), C_(2))
    l2, args = fa2pl.walk(func)
    assert l2(x=3) == 2
    assert args == {'x'}

    one = C_(lambda: 1)
    func = sub(add(S_('x'), one()), C_(2))
    l3, args = fa2pl.walk(func)
    assert l3(x=3) == 2
    assert args == {'x'}


def test_expression_walker_keep_fresh():
    s = S_.fresh()
    assert s.is_fresh
    walker = expression_walker.ExpressionWalker()
    result = walker.walk(s)
    assert result is s


def test_replace_expression_maintains_symbol_fresh():
    s = S_.fresh()
    assert s.is_fresh
    replacer = expression_walker.ReplaceExpressionWalker({})
    result = replacer.walk(s)
    assert result is s


def test_replace_expression_with_tuple():
    s1 = S_.fresh()
    s2 = S_.fresh()
    s3 = S_("s3")
    replacer = expression_walker.ReplaceExpressionWalker({s1: s3})
    result = replacer.walk((s2, (s1, s2)))
    assert isinstance(result, Tuple)
    assert result[0] is s2 and result[0].is_fresh
    assert isinstance(result[1], Tuple)
    assert result[1][0] == s3
    assert result[1][1] is s2 and result[1][1].is_fresh



import operator
import typing

import pytest

from ....datalog.expressions import Fact
from ....exceptions import (
    ForbiddenDisjunctionError,
    ForbiddenExpressionError,
    ProtectedKeywordError,
)
from ....expression_walker import ExpressionWalker
from ....expressions import Constant, Symbol
from ....logic import Conjunction, Implication, Union
from ...exceptions import (
    DistributionDoesNotSumToOneError,
    MalformedProbabilisticTupleError,
)
from ...expressions import (
    PROB,
    Condition,
    ProbabilisticPredicate,
    ProbabilisticQuery,
)
from ..program import CPLogicProgram, TranslateProbabilisticQueryMixin

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
Z = Symbol("Z")
Y = Symbol("Y")
p = Symbol("p")
p_1 = Symbol("p_1")
p_2 = Symbol("p_2")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")
a = Constant("a")
b = Constant("b")
c = Constant("c")
bernoulli = Symbol("bernoulli")


def test_probfact():
    probfact = Implication(
        ProbabilisticPredicate(Constant[float](0.2), P(x)),
        Constant[bool](True),
    )
    assert probfact.consequent.probability == Constant[float](0.2)
    assert probfact.consequent.body == P(x)


def test_deterministic_program():
    code = Union(
        (Implication(Z(x), Conjunction((P(x), Q(x)))), Fact(Q(a)), Fact(P(a)))
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    assert cpl_program.extensional_database() == {
        Q: Constant(frozenset({(a,)})),
        P: Constant(frozenset({(a,)})),
    }
    assert cpl_program.intensional_database() == {
        Z: Union((Implication(Z(x), Conjunction((P(x), Q(x)))),))
    }


def test_cplogic_program():
    cpl = CPLogicProgram()
    code = Union(
        [
            Implication(
                ProbabilisticPredicate(Constant[float](0.5), P(x)),
                Constant[bool](True),
            ),
            Implication(Q(x), Conjunction((P(x), Z(x)))),
            Fact(Z(a)),
            Fact(Z(b)),
        ]
    )
    cpl.walk(code)
    assert cpl.extensional_database() == {
        Z: Constant(frozenset({Constant((a,)), Constant((b,))}))
    }
    assert cpl.intensional_database() == {
        Q: Union((Implication(Q(x), Conjunction((P(x), Z(x)))),))
    }
    assert cpl.probabilistic_facts() == {
        P: Union(
            [
                Implication(
                    ProbabilisticPredicate(Constant[float](0.5), P(x)),
                    Constant[bool](True),
                )
            ]
        )
    }


def test_multiple_probfact_same_pred_symb():
    cpl = CPLogicProgram()
    code = Union(
        [
            Implication(
                ProbabilisticPredicate(Constant[float](0.5), P(a)),
                Constant[bool](True),
            ),
            Implication(
                ProbabilisticPredicate(Constant[float](0.2), P(b)),
                Constant[bool](True),
            ),
            Implication(Q(x), Conjunction((P(x), Z(x)))),
            Fact(Z(a)),
            Fact(Z(b)),
        ]
    )
    cpl.walk(code)
    assert cpl.extensional_database() == {
        Z: Constant(frozenset({Constant((a,)), Constant((b,))}))
    }
    assert cpl.intensional_database() == {
        Q: Union((Implication(Q(x), Conjunction((P(x), Z(x)))),))
    }
    assert len(cpl.probabilistic_facts()) == 1
    assert P in cpl.probabilistic_facts()
    probfacts = cpl.probabilistic_facts()[P]
    assert isinstance(probfacts, Constant[typing.AbstractSet])


def test_add_probfacts_from_tuple():
    cpl = CPLogicProgram()
    cpl.walk(Union(tuple()))
    cpl.add_probabilistic_facts_from_tuples(
        P, {(0.3, "hello", "gaston"), (0.7, "hello", "antonia")}
    )
    assert P in cpl.pfact_pred_symbs
    assert (
        Constant[float](0.7),
        Constant[str]("hello"),
        Constant[str]("antonia"),
    ) in cpl.symbol_table[P].value


def test_add_probfacts_from_tuple_no_probability():
    cpl = CPLogicProgram()
    cpl.walk(Union(tuple()))
    with pytest.raises(MalformedProbabilisticTupleError):
        cpl.add_probabilistic_facts_from_tuples(
            P, {("hello", "gaston"), ("hello", "antonia")}
        )


def test_add_probchoice_from_tuple():
    probchoice_as_tuples_iterable = {
        (0.5, "a", "a"),
        (0.2, "a", "b"),
        (0.3, "b", "b"),
    }
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(P, probchoice_as_tuples_iterable)
    assert P in cpl.symbol_table
    assert (
        Constant[float](0.2),
        Constant[str]("a"),
        Constant[str]("b"),
    ) in cpl.symbol_table[P].value


def test_add_probchoice_from_tuple_no_probability():
    cpl = CPLogicProgram()
    with pytest.raises(MalformedProbabilisticTupleError):
        cpl.add_probabilistic_choice_from_tuples(P, {("a", "b"), ("b", "b")})


def test_add_probchoice_from_tuple_twice_same_pred_symb():
    probchoice_as_tuples_iterable = {(1.0, "a", "a")}
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(P, probchoice_as_tuples_iterable)
    with pytest.raises(ForbiddenDisjunctionError):
        cpl.add_probabilistic_choice_from_tuples(
            P, probchoice_as_tuples_iterable
        )


def test_add_probchoice_does_not_sum_to_one():
    probchoice_as_tuples_iterable = {
        (0.5, "a", "a"),
        (0.2, "a", "b"),
        (0.1, "b", "b"),
    }
    cpl = CPLogicProgram()
    with pytest.raises(DistributionDoesNotSumToOneError):
        cpl.add_probabilistic_choice_from_tuples(
            P, probchoice_as_tuples_iterable
        )


def test_within_language_succ_query():
    rule = Implication(
        P(x, y, ProbabilisticQuery(PROB, (x, y))), Conjunction((Q(x), Z(y, z)))
    )
    cpl = CPLogicProgram()
    cpl.walk(rule)
    assert P in cpl.intensional_database()


def test_prob_protected_keyword():
    cpl = CPLogicProgram()
    with pytest.raises(ProtectedKeywordError):
        cpl.walk(Implication(PROB(x), Z(x)))


def test_within_language_succ_query_no_disjunction():
    q1 = Implication(
        P(x, ProbabilisticQuery(PROB, (x,))), Conjunction((Z(x), Q(x)))
    )
    q2 = Implication(
        P(x, ProbabilisticQuery(PROB, (x,))), Conjunction((Z(x), Y(y)))
    )
    cpl = CPLogicProgram()
    cpl.walk(q1)
    with pytest.raises(ForbiddenDisjunctionError):
        cpl.walk(q2)


def test_within_language_succ_query_invalid():
    q = Implication(
        P(x, y, ProbabilisticQuery(PROB, (x,))), Conjunction((Z(x), Q(y)))
    )
    cpl = CPLogicProgram()
    with pytest.raises(ForbiddenExpressionError):
        cpl.walk(q)
    q = Implication(
        P(x, ProbabilisticQuery(PROB, (x, y))), Conjunction((Z(x), Q(y)))
    )
    cpl = CPLogicProgram()
    with pytest.raises(ForbiddenExpressionError):
        cpl.walk(q)
    q = Implication(
        P(x, ProbabilisticQuery(PROB, (a, x))), Conjunction((Z(x), Q(y)))
    )
    cpl = CPLogicProgram()
    with pytest.raises(ForbiddenExpressionError):
        cpl.walk(q)


def test_wlq_marg_conditioned_conditioning_shared_var():
    """
    MARG task: Prob[ P(x) | Z(x) ] where variable x is shared by the
    conditioned predicate P(x) and by the conditioning predicate Z(x)
    """
    wlq = Implication(
        Q(x, ProbabilisticQuery(PROB, (x,))), Condition(P(x), Z(x))
    )
    cpl = CPLogicProgram()
    cpl.walk(wlq)
    assert Q in cpl.within_language_succ_queries()


def test_wlq_marg_conditioning_empty_conjunction():
    wlq = Implication(
        Q(x, ProbabilisticQuery(PROB, (x,))),
        Condition(P(x), Conjunction(tuple())),
    )
    cpl = CPLogicProgram()
    cpl.walk(wlq)
    assert Q in cpl.within_language_succ_queries()


def test_wlq_marg_conjunctive_conditioning():
    wlq = Implication(
        Q(x, y, ProbabilisticQuery(PROB, (x, y))),
        Condition(P(x), Conjunction((R(x), Q(y)))),
    )
    cpl = CPLogicProgram()
    cpl.walk(wlq)
    assert Q in cpl.within_language_succ_queries()


class _TestTranslator(TranslateProbabilisticQueryMixin, ExpressionWalker):
    pass


def test_wlq_floordiv_translation():
    wlq = Implication(
        Q(x, y, PROB(x, y)), Constant(operator.floordiv)(P(x), R(x, y))
    )
    translator = _TestTranslator()
    result = translator.walk(wlq)
    assert result == Implication(
        Q(x, y, ProbabilisticQuery(PROB, (x, y))), Condition(P(x), R(x, y))
    )


def test_wlq_marg_bad_syntax():
    bad_wlq = Implication(Q(x, y), Constant(operator.floordiv)(P(x), Z(x, y)))
    translator = _TestTranslator()
    with pytest.raises(ForbiddenExpressionError):
        translator.walk(bad_wlq)



from collections import OrderedDict
from typing import Iterable
from uuid import uuid1

import pandas as pd

from . import abstract as abc


class RelationalAlgebraStringExpression(str):
    def __repr__(self):
        return "{}{{ {} }}".format(self.__class__.__name__, super().__repr__())


class RelationalAlgebraColumn:
    pass


class RelationalAlgebraColumnInt(int, RelationalAlgebraColumn):
    pass


class RelationalAlgebraColumnStr(str, RelationalAlgebraColumn):
    pass


class RelationalAlgebraFrozenSet(abc.RelationalAlgebraFrozenSet):
    def __init__(self, iterable=None):
        self._container = None
        self._might_have_duplicates = True
        if iterable is not None:
            if isinstance(iterable, RelationalAlgebraFrozenSet):
                self._container = iterable._container
            elif isinstance(iterable, pd.DataFrame):
                self._container = iterable.copy()
            else:
                self._container = pd.DataFrame(iterable)

    def _drop_duplicates_if_needed(self):
        if self._might_have_duplicates:
            self._container = self._drop_duplicates(self._container)
            self._might_have_duplicates = False

    @classmethod
    def create_view_from(cls, other):
        if not isinstance(other, cls):
            raise ValueError(
                "View can only be created from an object of the same class"
            )
        output = cls()
        output._container = other._container
        return output

    @classmethod
    def dee(cls):
        output = cls()
        output._container = pd.DataFrame([()])
        return output

    @classmethod
    def dum(cls):
        return cls()

    def is_empty(self):
        return (
            self._container is None or
            len(self._container) == 0
        )

    def is_dum(self):
        return (
            self.arity == 0 and
            self.is_empty()
        )

    def is_dee(self):
        return (
            self.arity == 0 and
            not self.is_empty()
        )

    def __contains__(self, element):
        element = self._normalise_element(element)
        if (
            self.is_empty() or self.is_dee() or
            len(element) != self.arity
        ):
            res = False
        else:
            col = True
            for e, c in zip(element, self._container.iteritems()):
                col = col & (c[1] == e)
            res = col.any()
        return res

    @staticmethod
    def _normalise_element(element):
        if isinstance(element, tuple):
            pass
        elif hasattr(element, "__iter__"):
            element = tuple(element)
        else:
            element = (element, )
        return element

    def __iter__(self):
        if self.is_empty():
            values = {}
        elif self.is_dee():
            values = {tuple()}
        else:
            self._drop_duplicates_if_needed()
            values = self.itervalues()
        for v in values:
            yield tuple(v)

    def fetch_one(self):
        if self.is_dee():
            return tuple()
        return next(self._container.itertuples(name=None, index=False))

    def __len__(self):
        if self._container is None:
            return 0
        self._drop_duplicates_if_needed()
        return len(self._container)

    @staticmethod
    def _drop_duplicates(container):
        if len(container) == 0 or len(container.columns) == 0:
            return container
        container = container.drop_duplicates()
        return container

    @property
    def arity(self):
        if self._container is None:
            return 0
        return len(self._container.columns)

    @property
    def columns(self):
        return self._container.columns

    def as_numpy_array(self):
        res = self._container.values.view()
        res.setflags(write=False)
        return res

    def as_pandas_dataframe(self):
        return self._container

    def _empty_set_same_structure(self):
        return type(self)()

    def projection(self, *columns):
        if self.is_empty():
            return self._empty_set_same_structure()
        new_container = self._container[list(columns)]
        new_container.columns = pd.RangeIndex(len(columns))
        output = self._empty_set_same_structure()
        output._container = new_container
        return output

    def selection(self, select_criteria):
        if self.is_empty():
            return self._empty_set_same_structure()

        if callable(select_criteria):
            ix = self._container.apply(select_criteria, axis=1)
        elif isinstance(select_criteria, RelationalAlgebraStringExpression):
            ix = self._container.eval(select_criteria)
        else:
            ix = self._selection_dict(select_criteria)
        ix = ix.astype(bool)
        new_container = self._container[ix]

        output = self._empty_set_same_structure()
        output._container = new_container
        return output

    def _selection_dict(self, select_criteria):
        it = iter(select_criteria.items())
        col, value = next(it)
        ix = self._container[col] == value
        for col, value in it:
            if callable(value):
                selector = self._container[col].apply(value)
            else:
                selector = self._container[col] == value
            ix &= selector
        return ix

    def selection_columns(self, select_criteria):
        if self.is_empty():
            return self._empty_set_same_structure()
        it = iter(select_criteria.items())
        col1, col2 = next(it)
        ix = self._container[col1] == self._container[col2]
        for col1, col2 in it:
            ix &= self._container[col1] == self._container[col2]

        new_container = self._container[ix]

        output = self._empty_set_same_structure()
        output._container = new_container
        return output

    def equijoin(self, other, join_indices, return_mappings=False):
        res = self._dee_dum_product(other)
        if res is not None:
            return res
        if self.is_empty() or other.is_empty():
            return self._empty_set_same_structure()
        other_columns = range(
            self.arity, other._container.shape[1] + self.arity
        )
        ocont = other._container.copy(deep=False)
        ocont.columns = other_columns
        left_on, right_on = zip(*join_indices)
        left_on = list(left_on)
        right_on = list(l + self.arity for l in right_on)
        new_container = self._container.merge(
            ocont, left_on=left_on, right_on=right_on, sort=False
        )
        output = self._empty_set_same_structure()
        output._container = new_container
        output._might_have_duplicates = (
            self._might_have_duplicates |
            other._might_have_duplicates
        )
        return output

    def cross_product(self, other):
        res = self._dee_dum_product(other)
        if res is not None:
            return res
        if self.is_empty() or other.is_empty():
            return self._empty_set_same_structure()
        left = self._container.copy(deep=False)
        right = other._container.copy(deep=False)
        tmpcol = str(uuid1())
        left[tmpcol] = 1
        right[tmpcol] = 1
        new_container = pd.merge(left, right, on=tmpcol)
        del new_container[tmpcol]
        new_container.columns = range(new_container.shape[1])
        output = self._empty_set_same_structure()
        output._container = new_container
        output._might_have_duplicates = (
            self._might_have_duplicates |
            other._might_have_duplicates
        )
        return output

    def copy(self):
        output = self._empty_set_same_structure()
        if len(self) > 0:
            output._container = self._container.copy()
        return output

    def __repr__(self):
        if self.is_empty():
            return "{}"
        return repr(self._container.reset_index().drop("index", axis=1))

    def __or__(self, other):
        if self is other:
            return self.copy()
        elif isinstance(other, RelationalAlgebraFrozenSet):
            res = self._dee_dum_sum(other)
            if res is not None:
                return res
            ocont = other._container
            new_container = pd.merge(
                left=self._container,
                right=ocont,
                how="outer",
            )
            output = self._empty_set_same_structure()
            output._container = new_container
            return output
        else:
            return super().__or__(other)

    def __and__(self, other):
        if self is other:
            return self.copy()
        if isinstance(other, RelationalAlgebraFrozenSet):
            res = self._dee_dum_product(other)
            if res is not None:
                return res
            new_container = pd.merge(
                left=self._container,
                right=other._container,
                how="inner",
            )
            output = self._empty_set_same_structure()
            output._container = new_container
            output._might_have_duplicates = (
                self._might_have_duplicates |
                other._might_have_duplicates
            )
            return output
        else:
            return super().__and__(other)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            scont = self._container
            ocont = other._container
            if self.is_empty() and other.is_empty():
                res = True
            elif self.arity != other.arity:
                res = False
            elif self.arity == 0 and self.arity == 0:
                res = self.is_dee() and other.is_dee()
            elif scont is not None and ocont is not None:
                intersection_dups = scont.merge(
                    ocont, how='outer', indicator=True
                ).iloc[:, -1]
                res = (intersection_dups == 'both').all()
            else:
                res = False
            return res
        else:
            return super().__eq__(other)

    def groupby(self, columns):
        if not self.is_empty():
            if not isinstance(columns, Iterable):
                columns = [columns]
            for g_id, group in self._container.groupby(by=list(columns)):
                group_set = self._empty_set_same_structure()
                group_set._container = group
                yield g_id, group_set

    def itervalues(self):
        if self.is_empty():
            return iter([])
        else:
            return iter(self._container.itertuples(name=None, index=False))

    def __hash__(self):
        self._drop_duplicates_if_needed()
        v = self._container.values
        v.flags.writeable = False
        return hash(
            (tuple(self._container.columns), v.data.tobytes())
        )


class NamedRelationalAlgebraFrozenSet(
    RelationalAlgebraFrozenSet,
    abc.NamedRelationalAlgebraFrozenSet
):
    def __init__(self, columns, iterable=None):
        if isinstance(columns, NamedRelationalAlgebraFrozenSet):
            iterable = columns
            columns = columns.columns
        # ensure there is no duplicated column
        self._check_for_duplicated_columns(columns)
        self._columns = tuple(columns)
        self._might_have_duplicates = True
        if iterable is None:
            iterable = []

        if isinstance(iterable, NamedRelationalAlgebraFrozenSet):
            self._initialize_from_named_ra_set(iterable)
            self._might_have_duplicates = iterable._might_have_duplicates
        elif isinstance(iterable, RelationalAlgebraFrozenSet):
            self._initialize_from_unnamed_ra_set(iterable)
            self._might_have_duplicates = iterable._might_have_duplicates
        elif isinstance(iterable, pd.DataFrame):
            self._container = iterable.copy()
            self._container.columns = self._columns
        else:
            self._container = pd.DataFrame(
                iterable, columns=self._columns
            )

    def _initialize_from_named_ra_set(self, other):
        if (
            not other.is_dum()
            and self.arity != other.arity
        ):
            raise ValueError("Relations must have the same arity")

        if not other.is_empty():
            self._container = other._container[list(other.columns
                                                    )].copy(deep=False)
        else:
            self._container = pd.DataFrame(columns=self._columns)

    def _initialize_from_unnamed_ra_set(self, other):
        if other._container is None:
            self._container = pd.DataFrame(list(other), columns=self._columns)
        else:
            if self.arity != other.arity:
                raise ValueError("Relations must have the same arity")
            self._container = other._container.copy(deep=False)
            self._container.columns = self._columns

    @staticmethod
    def _check_for_duplicated_columns(columns):
        if len(set(columns)) != len(columns):
            columns = list(columns)
            dup_cols = set(c for c in columns if columns.count(c) > 1)
            raise ValueError(
                "Duplicated column names are not allowed. "
                f"Found the following duplicated columns: {dup_cols}"
            )

    @classmethod
    def create_view_from(cls, other):
        if not isinstance(other, cls):
            raise ValueError(
                "View can only be created from an object of the same class"
            )
        output = cls(columns=tuple())
        output._container = other._container
        output._columns = other._columns
        output._might_have_duplicates = other._might_have_duplicates
        return output

    def _empty_set_same_structure(self):
        return type(self)(self.columns)

    @classmethod
    def dee(cls):
        output = cls(())
        output._container = pd.DataFrame([()])
        return output

    @classmethod
    def dum(cls):
        return cls(())

    def _light_init_same_structure(
        self, container,
        might_have_duplicates=True,
        columns=None
    ):
        if columns is None:
            columns = self.columns
        output = type(self)(columns)
        output._container = container
        output._might_have_duplicates = might_have_duplicates
        return output

    @property
    def columns(self):
        return self._columns

    @property
    def arity(self):
        return len(self._columns)

    def __contains__(self, element):
        if isinstance(element, dict) and len(element) == self.arity:
            element = tuple(element[c] for c in self._container.columns)
        return super().__contains__(element)

    def projection(self, *columns):
        if self.is_empty():
            return type(self)(columns)
        if self.arity == 0:
            return self
        new_container = self._container[list(columns)]
        return self._light_init_same_structure(
            new_container,
            might_have_duplicates=True,
            columns=columns
        )

    def projection_to_unnamed(self, *columns):
        unnamed_self = self.to_unnamed()
        named_columns = list(self.columns)
        columns = tuple(named_columns.index(c) for c in columns)
        return unnamed_self.projection(*columns)

    def equijoin(self, other, join_indices, return_mappings=False):
        raise NotImplementedError()

    def naturaljoin(self, other):
        res = self._dee_dum_product(other)
        if res is not None:
            return res
        on = [c for c in self.columns if c in other.columns]

        if len(on) == 0:
            return self.cross_product(other)

        new_columns = self.columns + tuple(
            c for c in other.columns if c not in self.columns
        )

        new_container = self._container.merge(other._container)
        return self._light_init_same_structure(
            new_container,
            might_have_duplicates=(
                self._might_have_duplicates |
                other._might_have_duplicates
            ),
            columns=new_columns
        )

    def cross_product(self, other):
        res = self._dee_dum_product(other)
        if res is not None:
            return res.copy()
        if len(self._container.columns.intersection(other.columns)) > 0:
            raise ValueError(
                "Cross product with common columns "
                "is not valid"
            )

        new_columns = self.columns + other.columns
        if self.is_empty() or other.is_empty():
            res = type(self)(new_columns)
        else:
            left = self._container.copy(deep=False)
            right = other._container.copy(deep=False)
            tmpcol = str(uuid1())
            left[tmpcol] = 1
            right[tmpcol] = 1
            new_container = pd.merge(left, right, on=tmpcol)
            del new_container[tmpcol]
            new_container.columns = (
                tuple(self._container.columns) +
                tuple(other._container.columns)
            )
            res = self._light_init_same_structure(
                new_container,
                might_have_duplicates=(
                    self._might_have_duplicates |
                    other._might_have_duplicates
                ),
                columns=new_columns
            )
        return res

    def rename_column(self, src, dst):
        if src not in self._columns:
            raise ValueError(f"{src} not in columns")
        if src == dst:
            return self
        if dst in self._columns:
            raise ValueError(f"{dst} cannot be in the columns")
        src_idx = self._columns.index(src)
        new_columns = self._columns[:src_idx] + (dst,
                                                 ) + self._columns[src_idx +
                                                                   1:]
        new_container = self._container.rename(columns={src: dst})
        return self._light_init_same_structure(
            new_container,
            might_have_duplicates=self._might_have_duplicates,
            columns=new_columns
        )

    def rename_columns(self, renames):
        # prevent duplicated destination columns
        self._check_for_duplicated_columns(renames.values())
        if not set(renames).issubset(self.columns):
            # get the missing source columns
            # for a more convenient error message
            not_found_cols = set(c for c in renames if c not in self._columns)
            raise ValueError(
                f"Cannot rename non-existing columns: {not_found_cols}"
            )
        new_columns = tuple(
            renames.get(col, col) for col in self._columns
        )
        new_container = self._container.rename(columns=renames)
        return self._light_init_same_structure(
            new_container,
            might_have_duplicates=self._might_have_duplicates,
            columns=new_columns
        )

    def __eq__(self, other):
        scont = self._container
        ocont = other._container
        if len(scont.columns.symmetric_difference(ocont.columns)) > 0:
            res = False
        elif len(scont) == 0 and len(ocont) == 0:
            res = True
        elif len(scont.columns) == 0 and len(ocont.columns) == 0:
            res = len(scont) > 0 and len(ocont) > 0
        else:
            intersection_dups = scont.merge(
                ocont, how='outer', indicator=True
            ).iloc[:, -1]
            res = (intersection_dups == 'both').all()
        return res

    def groupby(self, columns):
        if self.is_empty():
            return
        for g_id, group in self._container.groupby(by=list(columns)):
            group_set = self._light_init_same_structure(
                group,
                might_have_duplicates=self._might_have_duplicates,
                columns=self.columns
            )
            yield g_id, group_set

    def aggregate(self, group_columns, aggregate_function):
        group_columns = list(group_columns)
        if len(set(group_columns)) < len(group_columns):
            raise ValueError("Cannot group on repeated columns")
        self._drop_duplicates_if_needed()
        if len(group_columns) > 0:
            groups = self._container.groupby(group_columns)
        else:
            groups = self._container.groupby(lambda x: 0)

        aggs, aggs_multi_column = self._classify_aggregations(
            group_columns, aggregate_function
        )

        new_containers = []
        if len(aggs) > 0:
            new_containers.append(groups.agg(**aggs))

        for dst, fun in aggs_multi_column.items():
            new_col = (
                groups
                .apply(fun)
                .rename(dst)
                .to_frame()
            )
            new_containers.append(new_col)

        new_container = pd.concat(new_containers, axis=1)

        if len(group_columns) > 0:
            new_container = new_container.reset_index()

        self._keep_column_types(
            new_container, set(aggs) |
            set(aggs_multi_column)
        )

        output = self._light_init_same_structure(
            new_container,
            might_have_duplicates=self._might_have_duplicates,
            columns=list(new_container.columns)
        )
        return output

    def _keep_column_types(self, new_container, skip=None):
        if self.is_empty():
            return

        if skip is None:
            skip = {}
        for col in new_container.columns:
            if col in skip:
                continue
            if (
                col in self._container.columns and
                new_container[col].dtype != self._container[col].dtype
            ):
                new_container[col] = new_container[col].astype(
                    self._container[col].dtype
                )

    def _classify_aggregations(self, group_columns, aggregate_function):
        aggs = OrderedDict()
        aggs_multi_columns = OrderedDict()
        if isinstance(aggregate_function, dict):
            arg_iterable = (
                (k, k if k in self.columns else None, v)
                for k, v in aggregate_function.items()
            )
        elif isinstance(aggregate_function, (tuple, list)):
            arg_iterable = aggregate_function
        else:
            raise ValueError(
                "Unsupported aggregate_function: {} of type {}".format(
                    aggregate_function, type(aggregate_function)
                )
            )

        for dst, src, fun in arg_iterable:
            if dst in group_columns:
                raise ValueError(
                    f"Destination column {dst} can't be part of the grouping"
                )
            if src in self.columns:
                aggs[dst] = pd.NamedAgg(src, fun)
            elif src is None or all(s in self.columns for s in src):
                aggs_multi_columns[dst] = fun
            else:
                raise ValueError(f"Source column {src} not in columns")

        return aggs, aggs_multi_columns

    def extended_projection(self, eval_expressions):
        proj_columns = list(eval_expressions.keys())
        if self.is_empty():
            return NamedRelationalAlgebraFrozenSet(
                columns=proj_columns, iterable=[],
            )
        new_container = self._container.copy()
        for dst_column, operation in eval_expressions.items():
            if isinstance(operation, RelationalAlgebraStringExpression):
                if str(operation) != str(dst_column):
                    new_container = new_container.eval(
                        "{}={}".format(str(dst_column), str(operation)),
                        engine='python'
                    )
            elif isinstance(operation, RelationalAlgebraColumn):
                new_container[dst_column] = new_container[operation]
            elif callable(operation):
                new_container[dst_column] = new_container.apply(
                    operation, axis=1
                )
            else:
                new_container[dst_column] = operation
        new_container = new_container[proj_columns]
        output = self._light_init_same_structure(
            new_container,
            might_have_duplicates=self._might_have_duplicates,
            columns=proj_columns
        )
        return output

    def __iter__(self):
        self._drop_duplicates_if_needed()
        if self.is_dee():
            return iter([tuple()])
        container = self._container[list(self.columns)]
        return container.itertuples(index=False, name="tuple")

    def fetch_one(self):
        if self.is_dee():
            return tuple()
        container = self._container[list(self.columns)]
        return next(container.itertuples(index=False, name="tuple"))

    def to_unnamed(self):
        container = self._container[list(self.columns)].copy()
        container.columns = range(len(container.columns))
        output = RelationalAlgebraFrozenSet()
        output._container = container
        return output

    def __sub__(self, other):
        if (
            (self.arity > 0 and other.arity > 0) and
            len(
                self._container.columns.difference(
                    other._container.columns
                )
            ) > 0
        ):
            raise ValueError(
                "Difference defined only for sets with the same columns"
            )
        if self.is_empty() or other.is_empty():
            return self.copy()
        if self.is_dee():
            if other.is_dee():
                return self.dum()
            return self.dee()
        new_container = self._container.merge(
            other._container,
            indicator=True,
            how='left'
        )
        new_container = new_container[
            new_container.iloc[:, -1] == 'left_only'
        ].iloc[:, :-1]

        self._keep_column_types(new_container)
        output = self._light_init_same_structure(
            new_container,
            might_have_duplicates=self._might_have_duplicates,
        )
        return output

    def __or__(self, other):
        res = self._dee_dum_sum(other)
        if res is not None:
            return res
        elif set(self.columns) != set(other.columns):
            raise ValueError(
                "Union defined only for sets with the same columns"
            )
        new_container = pd.merge(
            left=self._container,
            right=other._container,
            how="outer",
        )

        self._keep_column_types(new_container)
        output = self._light_init_same_structure(
            new_container,
            might_have_duplicates=True,
        )
        return output

    def __and__(self, other):
        res = self._dee_dum_product(other)
        if res is not None:
            return res
        if set(self.columns) != set(other.columns):
            raise ValueError(
                "Union defined only for sets with the same columns"
            )
        if self.is_empty():
            return self.copy()
        new_container = pd.merge(
            left=self._container,
            right=other._container,
            how="inner",
        )
        self._keep_column_types(new_container)
        output = self._light_init_same_structure(
            new_container,
            might_have_duplicates=self._might_have_duplicates,
        )
        return output

    def __lt__(self, other):
        raise NotImplementedError()

    def __le__(self, other):
        raise NotImplementedError()

    def __gt__(self, other):
        raise NotImplementedError()

    def __ge__(self, other):
        raise NotImplementedError()


class RelationalAlgebraSet(
    RelationalAlgebraFrozenSet,
    abc.RelationalAlgebraSet
):
    def add(self, value):
        value = self._normalise_element(value)
        e_hash = hash(value)
        if self.is_empty():
            self._container = pd.DataFrame([value], index=[e_hash])
        else:
            self._container.loc[e_hash] = value

    def discard(self, value):
        if not self.is_empty():
            try:
                value = self._normalise_element(value)
                col = True
                for e, c in zip(value, self._container.iteritems()):
                    col = col & (c[1] == e)
                ix = self._container.index[col]
                self._container.drop(index=ix, inplace=True)
            except KeyError:
                pass

    def __ior__(self, other):
        if isinstance(other, RelationalAlgebraSet):
            if other.is_empty() or other.arity == 0:
                return self
            if self.is_empty():
                self._container = other._container.copy()
                self._might_have_duplicates = other._might_have_duplicates
                return self
            if other.arity != self.arity:
                raise ValueError(
                    "Operation only valid for sets with the same arity"
                )
            new_container = pd.merge(
                left=self._container,
                right=other._container,
                how="outer",
            )
            self._might_have_duplicates = True
            self._container = new_container
            return self
        else:
            return super().__ior__(other)

    def __isub__(self, other):
        if self.is_empty():
            return self
        if isinstance(other, RelationalAlgebraSet):
            if other.is_empty() or other.arity == 0:
                if self.is_empty() or self.arity == 0:
                    self._container = self._container.iloc[:0]
            elif other.arity != self.arity:
                raise ValueError(
                    "Operation only valid for sets with the same arity"
                )
            else:
                new_container = pd.merge(
                    left=self._container,
                    right=other._container,
                    how="left",  indicator=True
                )
                new_container = new_container[
                    new_container.iloc[:, -1] == 'left_only'
                ].iloc[:, :-1]
                self._container = new_container
            return self
        else:
            return super().__isub__(other)



from neurolang.exceptions import ForbiddenUnstratifiedAggregation
from typing import AbstractSet

import pytest

from ...expression_walker import ExpressionBasicEvaluator
from ...expressions import (Constant, ExpressionBlock, NeuroLangException,
                            Symbol)
from ...type_system import Unknown
from .. import DatalogProgram, Fact, Implication
from ..aggregation import (AggregationApplication, Chase,
                           DatalogWithAggregationMixin,
                           TranslateToLogicWithAggregation)
from ..expressions import Union

S_ = Symbol
C_ = Constant
Imp_ = Implication
Fa_ = AggregationApplication
Eb_ = ExpressionBlock
U_ = Union
F_ = Fact


class Datalog(
    TranslateToLogicWithAggregation,
    DatalogWithAggregationMixin, DatalogProgram,
    ExpressionBasicEvaluator
):
    def function_sum(self, x: AbstractSet) -> Unknown:
        return sum(v for v in x)

    def function_sum2(self, x: AbstractSet, y: AbstractSet) -> Unknown:
        return sum(v + w for v, w in zip(x, y))

    def function_set_create(self, x: AbstractSet) -> Unknown:
        return frozenset(x)


def test_aggregation_parsing():
    dl = Datalog()

    P = S_('P')  # noqa: N806
    Q = S_('Q')  # noqa: N806
    x = S_('x')
    y = S_('y')

    edb = [
        F_(P(C_(i), C_(j)))
        for i in range(3)
        for j in range(3)
    ]

    code = U_(edb + [
        Imp_(Q(x, S_('sum')(y,)), P(x, y)),
    ])

    dl.walk(code)

    assert Q in dl.intensional_database()
    assert P in dl.extensional_database()

    with pytest.raises(NeuroLangException):
        dl.walk(U_([
            Imp_(Q(x, Fa_(C_(sum), (y,)), Fa_(C_(sum), (y,))), P(x, y)),
        ]))


def test_aggregation_non_stratified():
    P = S_('P')  # noqa: N806
    Q = S_('Q')  # noqa: N806
    x = S_('x')

    edb = [
        F_(P(C_(i)))
        for i in range(3)
    ]

    code = Eb_(edb + [
        Imp_(Q(Fa_(S_('sum'), (x,))), Q(x)),
    ])

    dl = Datalog()
    dl.walk(code)

    chase = Chase(dl)

    with pytest.raises(ForbiddenUnstratifiedAggregation):
        chase.build_chase_solution()


def test_aggregation_chase_no_grouping():

    P = S_('P')  # noqa: N806
    Q = S_('Q')  # noqa: N806
    R = S_('R')  # noqa: N806
    x = S_('x')

    edb = [
        F_(P(C_(i)))
        for i in range(3)
    ]

    code = Eb_(edb + [
        Imp_(Q(Fa_(S_('sum'), (x,))), P(x)),
    ])

    dl = Datalog()
    dl.walk(code)

    chase = Chase(dl)

    solution = chase.build_chase_solution()

    dl.add_extensional_predicate_from_tuples(R, {(3,)})

    res = dl.extensional_database()['R']

    assert solution[Q] == res


def test_aggregation_chase_no_grouping_2args():

    P = S_('P')  # noqa: N806
    Q = S_('Q')  # noqa: N806
    R = S_('R')  # noqa: N806
    x = S_('x')
    y = S_('y')

    edb = [
        F_(P(C_(i), C_(2 * i)))
        for i in range(3)
    ]

    code = U_(edb + [
        Imp_(Q(Fa_(S_('sum2'), (x, y))), P(x, y)),
    ])

    dl = Datalog()
    dl.walk(code)

    chase = Chase(dl)

    solution = chase.build_chase_solution()

    dl.add_extensional_predicate_from_tuples(R, {(9,)})

    res = dl.extensional_database()['R']

    assert solution[Q] == res


def test_aggregation_chase_single_grouping():
    dl = Datalog()

    P = S_('P')  # noqa: N806
    Q = S_('Q')  # noqa: N806
    R = S_('R')  # noqa: N806
    x = S_('x')
    y = S_('y')

    edb = [
        F_(P(C_(i), C_(i * j)))
        for i in range(3)
        for j in range(3)
    ]

    code = Eb_(edb + [
        Imp_(Q(x, Fa_(S_('sum'), (y,))), P(x, y)),
    ])

    dl.walk(code)

    chase = Chase(dl)

    solution = chase.build_chase_solution()

    dl.add_extensional_predicate_from_tuples(
        R,
        {
            (0, 0),
            (1, 3),
            (2, 6)
        }
    )
    res = dl.extensional_database()['R']

    assert solution[Q] == res


def test_aggregation_chase_single_grouping_muliple_columns():
    dl = Datalog()

    P = S_('P')  # noqa: N806
    Q = S_('Q')  # noqa: N806
    R = S_('R')  # noqa: N806
    x = S_('x')
    y = S_('y')
    z = S_('z')

    edb = [
        F_(P(C_(i), C_(i * j), C_(0)))
        for i in range(3)
        for j in range(3)
    ]

    code = Eb_(edb + [
        Imp_(Q(x, Fa_(S_('sum2'), (y, z))), P(x, y, z)),
    ])

    dl.walk(code)

    chase = Chase(dl)

    solution = chase.build_chase_solution()

    dl.add_extensional_predicate_from_tuples(
        R,
        {
            (0, 0),
            (1, 3),
            (2, 6)
        }
    )
    res = dl.extensional_database()['R']

    assert solution[Q] == res


def test_aggregation_set_creation():
    dl = Datalog()
    P = S_('P')
    Q = S_('Q')
    x = S_('x')
    y = S_('y')

    edb = tuple(
        F_(P(C_(0), C_(i)))
        for i in range(3)
    )

    code = Eb_(edb + (
        Imp_(
            Q(x, Fa_(S_('set_create'), (y,))),
            P(x, y)
        ),
    ))

    dl.walk(code)
    chase = Chase(dl)
    solution = chase.build_chase_solution()

    assert Q in solution
    assert set(solution[Q].value.unwrapped_iter()) == {
        (0, frozenset(i for i in range(3)))
    }


def test_aggregation_emptyset():
    dl = Datalog()

    P = S_('P')  # noqa: N806
    Q = S_('Q')  # noqa: N806
    R = S_('R')  # noqa: N806
    x = S_('x')
    y = S_('y')

    edb = [
        F_(P(C_(i), C_(i * j)))
        for i in range(3)
        for j in range(3)
    ] + [F_(R(C_(10)))]

    code = Eb_(edb + [
        Imp_(Q(x, Fa_(S_('sum'), (y,))), R(x) & P(x, y)),
    ])

    dl.walk(code)

    chase = Chase(dl)

    solution = chase.build_chase_solution()

    assert Q not in solution or solution[Q] == set()



from ... import expressions
from ...logic import Conjunction, Disjunction, Union
from ..unification import (
    apply_substitution,
    merge_substitutions,
    most_general_unifier,
)

C_ = expressions.Constant
S_ = expressions.Symbol
F_ = expressions.FunctionApplication


def test_unification():
    a = S_("a")
    b = S_("b")
    x = S_("x")
    y = S_("y")
    z = S_("z")

    assert most_general_unifier(a(x), a(x)) == (dict(), a(x))
    assert most_general_unifier(a(x), a(y)) == (dict(x=y), a(y))
    assert most_general_unifier(a(x, y), a(y, y)) == (dict(x=y), a(y, y))
    assert most_general_unifier(a(x, z), a(y, y)) == (dict(x=y, z=y), a(y, y))
    assert most_general_unifier(a(C_(1), z), a(y, y)) == (
        dict(y=C_(1), z=C_(1)),
        a(C_(1), C_(1)),
    )
    assert most_general_unifier(a(C_(1), C_(2)), a(y, y)) is None
    assert most_general_unifier(a(b(x)), a(b(y))) == (dict(x=y), a(b(y)))
    assert most_general_unifier(a(b(x), z), a(b(y), y)) == (
        dict(x=y, z=y),
        a(b(y), y),
    )

    assert most_general_unifier(a(x, y), a(x)) is None
    assert most_general_unifier(a(b(x), y), a(x, y)) is None


def test_apply_substitution_naryoperator():
    P = S_("P")
    Q = S_("Q")
    x = S_("x")
    y = S_("y")
    z = S_("z")

    union = Union((P(x), Q(x)))
    expected = Union((P(y), Q(y)))
    assert apply_substitution(union, {x: y}) == expected
    assert apply_substitution(union, {y: x}) == union

    union = Union((P(y), Q(x)))
    expected_1 = Union((P(y), Q(y)))
    expected_2 = Union((P(x), Q(x)))

    assert apply_substitution(union, {x: y}) == expected_1
    assert apply_substitution(union, {y: x}) == expected_2

    union = Union((P(y), Q(x, z)))
    expected_1 = Union((P(y), Q(y, z)))
    expected_2 = Union((P(x), Q(x, z)))

    assert apply_substitution(union, {x: y}) == expected_1
    assert apply_substitution(union, {y: x}) == expected_2

    union = Conjunction((P(x), Q(x)))
    expected = Conjunction((P(y), Q(y)))
    assert apply_substitution(union, {x: y}) == expected
    assert apply_substitution(union, {y: x}) == union

    union = Conjunction((P(y), Q(x)))
    expected_1 = Conjunction((P(y), Q(y)))
    expected_2 = Conjunction((P(x), Q(x)))

    assert apply_substitution(union, {x: y}) == expected_1
    assert apply_substitution(union, {y: x}) == expected_2

    union = Conjunction((P(y), Q(x, z)))
    expected_1 = Conjunction((P(y), Q(y, z)))
    expected_2 = Conjunction((P(x), Q(x, z)))

    assert apply_substitution(union, {x: y}) == expected_1
    assert apply_substitution(union, {y: x}) == expected_2

    union = Disjunction((P(x), Q(x)))
    expected = Disjunction((P(y), Q(y)))
    assert apply_substitution(union, {x: y}) == expected
    assert apply_substitution(union, {y: x}) == union

    union = Disjunction((P(y), Q(x)))
    expected_1 = Disjunction((P(y), Q(y)))
    expected_2 = Disjunction((P(x), Q(x)))

    assert apply_substitution(union, {x: y}) == expected_1
    assert apply_substitution(union, {y: x}) == expected_2

    union = Disjunction((P(y), Q(x, z)))
    expected_1 = Disjunction((P(y), Q(y, z)))
    expected_2 = Disjunction((P(x), Q(x, z)))

    assert apply_substitution(union, {x: y}) == expected_1
    assert apply_substitution(union, {y: x}) == expected_2


def test_merge_substitutions():
    assert merge_substitutions({1: 2}, {3: 4}) == {1: 2, 3: 4}
    assert merge_substitutions({1: 2}, {1: 2}) == {1: 2}
    assert merge_substitutions({1: 2}, {1: 4}) is None



import collections
from itertools import chain

from . import expressions
from .exceptions import NeuroLangException


__all__ = ['TypedSymbolTable']


class TypedSymbolTable(collections.abc.MutableMapping):
    def __init__(self, enclosing_scope=None, readonly=False):
        self._symbols = collections.OrderedDict()

        self._symbols_by_type = collections.defaultdict(dict)
        self.enclosing_scope = enclosing_scope
        self.readonly = readonly

    def __len__(self):
        return len(self._symbols)

    def __getitem__(self, key):
        try:
            return self._symbols[key]
        except KeyError:
            if self.enclosing_scope is not None:
                return self.enclosing_scope[key]
            else:
                raise KeyError("Expression %s not in the table" % key)

    def __setitem__(self, key, value):
        if self.readonly:
            raise NeuroLangException("This symbol table is readonly")
        if isinstance(value, expressions.Expression):
            self._symbols[key] = value
            if value.type not in self._symbols_by_type:
                self._symbols_by_type[value.type] = dict()
            self._symbols_by_type[value.type][key] = value
        elif value is None:
            self._symbols[key] = None
        else:
            raise ValueError("Wrong assignment %s" % str(value))

    def __delitem__(self, key):
        value = self._symbols[key]
        del self._symbols_by_type[value.type][key]
        del self._symbols[key]

    def __iter__(self):
        keys = iter(self._symbols.keys())
        if self.enclosing_scope is not None:
            keys = chain(keys, iter(self.enclosing_scope))

        return keys

    def __repr__(self):
        return '{%s}' % (
            ', '.join([
                '%s: (%s)' % (k, v)
                for k, v in self._symbols.items()
            ])
        )

    def types(self):
        ret = self._symbols_by_type.keys()
        if self.enclosing_scope is not None:
            ret = ret | self.enclosing_scope.types()
        return ret

    def symbols_by_type(self, type_, include_subtypes=True):
        ret = dict()
        if self.enclosing_scope is not None:
            ret.update(self.enclosing_scope.symbols_by_type(type_))

        for t in self.types():
            if (
                t is type_ or (
                    include_subtypes and
                    t is not expressions.Unknown and
                    expressions.is_leq_informative(t, type_)
                )
            ):
                ret.update(self._symbols_by_type[t])

        return ret

    def create_scope(self):
        subscope = TypedSymbolTable(enclosing_scope=self)
        return subscope

    def set_readonly(self, readonly):
        self.readonly = readonly



import numpy as np
from nilearn import datasets

import nibabel as nib
from neurolang import frontend as fe


class TimeRegionComparisons:
    params = [
        [
            'anterior_of', 'posterior_of', 'overlapping',
            'superior_of', 'inferior_of', 'left_of', 'right_of'
        ]
    ]

    param_names = ['direction']

    timeout = 10 * 60

    def setup(self, direction):
        raise NotImplementedError("Skip benchmark")
        if not hasattr(self, 'nl'):
            self.setup_()

    def setup_(self):
        nl = fe.NeurolangDL()

        destrieux_atlas = datasets.fetch_atlas_destrieux_2009()
        yeo_atlas = datasets.fetch_atlas_yeo_2011()

        img = nib.load(destrieux_atlas['maps'])
        aff = img.affine
        data = img.get_data()
        rset = []
        for label, name in destrieux_atlas['labels']:
            if label == 0:
                continue
            voxels = np.transpose((data == label).nonzero())
            if len(voxels) == 0:
                continue
            rset.append((
                name.decode('utf8'),
                fe.ExplicitVBR(
                    voxels, aff,
                    image_dim=img.shape, prebuild_tree=True
                )
            ))
        nl.add_tuple_set(rset, name='destrieux')

        img = nib.load(yeo_atlas['thick_17'])
        aff = img.affine
        data = img.get_data().squeeze()
        rset = []
        for label in range(1, 18):
            name = str(label)
            if label == 0:
                continue
            voxels = np.transpose((data == label).nonzero())
            if len(voxels) == 0:
                continue
            rset.append((
                name,
                fe.ExplicitVBR(
                    voxels, aff,
                    image_dim=data.shape,
                    prebuild_tree=True
                )
            ))
        nl.add_tuple_set(rset, name='yeo')
        self.nl = nl

    def time_spatial_relation(self, direction):
        with self.nl.scope as e:
            self.nl.query(
                (e.d,),
                e.destrieux(e.d, e.dr) & e.yeo('1', e.dy) &
                e[direction](e.dr, e.dy)
            )



"""
Support for aggregations according to [1]_. For instance
the encoding of ``P(x, count<y>):-Q(x, y)`` in intermediate
representation ``Implication(P(x, AggregationApplication(count, y), Q(x, y))``
produces the set ``P(x, z)`` where ``z`` is the number of tuples ``(x, ...)``
in the set ``Q``.

.. [1] T. J. Green, S. S. Huang, B. T. Loo, W. Zhou,
   Datalog and Recursive Query Processing.
   FNT in Databases. 5, 105–195 (2012).
"""

from warnings import warn

from ..exceptions import ForbiddenUnstratifiedAggregation, NeuroLangException
from ..expression_walker import (
    FunctionApplicationToPythonLambda,
    PatternWalker,
    ReplaceSymbolsByConstants,
    add_match
)
from ..expressions import Constant, Expression, FunctionApplication, Symbol
from ..type_system import get_generic_type
from ..utils.relational_algebra_set import RelationalAlgebraStringExpression
from . import (
    Implication,
    Union,
    chase,
    is_conjunctive_expression_with_nested_predicates
)
from .basic_representation import UnionOfConjunctiveQueries
from .expression_processing import extract_logic_predicates, stratify
from .expressions import TranslateToLogic
from .instance import MapInstance

FA2L = FunctionApplicationToPythonLambda()


class AggregationApplication(FunctionApplication):
    def __repr__(self):
        r = u'\u03BB{{<{}>: {}}}'.format(self.functor, self.__type_repr__)
        if self.args is ...:
            r += '(...)'
        elif self.args is not None:
            r += (
                '(' +
                ', '.join(repr(arg) for arg in self.args)
                + ')'
                )

        return r


def is_aggregation_rule(rule):
    return is_aggregation_predicate(rule.consequent)


def is_aggregation_predicate(predicate):
    return any(
        isinstance(arg, AggregationApplication) for arg in predicate.args
    )


class TranslateToLogicWithAggregation(TranslateToLogic):
    @add_match(
        Implication(FunctionApplication(Symbol, ...), Expression),
        lambda rule: any(
            isinstance(arg, FunctionApplication)
            and get_generic_type(type(arg)) is FunctionApplication
            for arg in rule.consequent.args
        )
    )
    def transform_function_application_consequent_to_aggregation(self, rule):
        consequent_arguments = tuple()
        for arg in rule.consequent.args:
            if (
                isinstance(arg, FunctionApplication)
                and get_generic_type(type(arg)) is FunctionApplication
            ):
                arg = AggregationApplication(*arg.unapply())
            consequent_arguments += (arg,)

        return self.walk(
            Implication(
                rule.consequent.functor(*consequent_arguments),
                rule.antecedent
            )
        )


class DatalogWithAggregationMixin(PatternWalker):
    @add_match(
        Implication(FunctionApplication(Symbol, ...), Expression),
        is_aggregation_rule
    )
    def statement_intensional_aggregation(self, expression):
        consequent = expression.consequent
        antecedent = expression.antecedent

        self._validate_aggregation_implication_syntax(consequent, antecedent)

        if consequent.functor in self.symbol_table:
            eb = self._new_intensional_internal_representation(consequent)
        else:
            eb = tuple()

        eb = eb + (expression, )
        symbol = consequent.functor.cast(UnionOfConjunctiveQueries)
        self.symbol_table[symbol] = Union(eb)

        return expression

    def _validate_aggregation_implication_syntax(self, consequent, antecedent):
        if consequent.functor in self.protected_keywords:
            raise NeuroLangException(
                f'symbol {self.constant_set_name} is protected'
            )

        seen_aggregations = 0
        for arg in consequent.args:
            if isinstance(arg, AggregationApplication):
                seen_aggregations += 1
                aggregation_functor = arg.functor
            elif not isinstance(arg, (Constant, Symbol)):
                raise NeuroLangException(
                    f'The consequent {consequent} can only be '
                    'constants, symbols'
                )

            if seen_aggregations > 1:
                raise NeuroLangException(
                    f'Only one aggregation allowed in {consequent}'
                )

        consequent_symbols = (
            consequent._symbols - consequent.functor._symbols -
            aggregation_functor._symbols
        )

        if not consequent_symbols.issubset(antecedent._symbols):
            raise NeuroLangException(
                "All variables on the consequent need to be on the antecedent"
            )

        if not is_conjunctive_expression_with_nested_predicates(antecedent):
            raise NeuroLangException(
                f'Expression {antecedent} is not conjunctive'
            )


class Chase(chase.Chase):
    def check_constraints(self, instance_update):
        code = Union(tuple(self.rules))
        stratified_code, stratifiable = stratify(code, self.datalog_program)
        self.stratified_code = stratified_code

        for stratum in self.stratified_code:
            seen_in_stratum = set()
            aggregate_rules = []
            for rule in stratum:
                seen_in_stratum.add(rule.consequent.functor)
                if is_aggregation_rule(rule):
                    aggregate_rules.append(rule)

            self._stratum_is_aggregation_viable(
                seen_in_stratum, aggregate_rules
            )

        return super().check_constraints(instance_update)

    def build_chase_solution(self):
        instance_update = MapInstance(
            self.datalog_program.extensional_database()
        )
        self.check_constraints(instance_update)
        instance = MapInstance()
        for stratum in self.stratified_code:
            instance = self.execute_chase(stratum, instance_update, instance)
            instance_update = instance
            instance = MapInstance()
        return instance_update

    def _stratum_is_aggregation_viable(self, seen_in_stratum, aggregate_rules):
        for rule in aggregate_rules:
            if any(
                predicate.functor in seen_in_stratum
                for predicate
                in extract_logic_predicates(rule.antecedent)
            ):
                raise ForbiddenUnstratifiedAggregation(
                    f"Unstratifiable aggregation {rule.consequent.functor}"
                )

    def compute_result_set(
        self, rule, substitutions, instance, restriction_instance=None
    ):
        if not is_aggregation_rule(rule):
            return super().compute_result_set(
                rule,
                substitutions,
                instance,
                restriction_instance=restriction_instance
            )
        if restriction_instance is None:
            restriction_instance = dict()

        group_vars, output_args = self._get_groups_and_aggregations(rule)
        new_tuples = (
            substitutions
            .aggregate(group_vars, output_args)
            .projection_to_unnamed(*(oa[0] for oa in output_args))
        )
        new_tuples = self.datalog_program.new_set(new_tuples)
        return self.compute_instance_update(
            rule, new_tuples, instance, restriction_instance
        )

    def _get_groups_and_aggregations(self, rule):
        output_args = []
        group_vars = []
        for arg in rule.consequent.args:
            if isinstance(arg, Symbol):
                group_vars.append(arg.name)
                aggregation_args = arg.name
                fun = RelationalAlgebraStringExpression('first')
            elif isinstance(arg, AggregationApplication):
                aggregation_args, fun = self._obtain_aggregations(arg)
            output_args.append(
                (
                    Symbol.fresh().name,
                    aggregation_args,
                    fun
                )
            )
        return group_vars, output_args

    def _obtain_aggregations(self, arg):
        if (
            len(arg.args) == 1 and
            isinstance(arg.args[0], Symbol)
        ):
            fun = self.datalog_program.walk(arg.functor).value
            aggregation_args = arg.args[0].name
        else:
            aggregation_args = None
            fa = ReplaceSymbolsByConstants(
                self.datalog_program.symbol_table
            ).walk(FunctionApplication(*arg.unapply()))
            fun_, arg_names = FA2L.walk(fa)
            fun_str = (
                "lambda t: fun_(" +
                ", ".join(f't.{arg_}' for arg_ in arg_names) +
                ")"
            )
            gs = globals()
            ls = locals()
            gs['fun_'] = fun_
            fun = eval(fun_str, gs, ls)
        return aggregation_args, fun

    def eliminate_already_computed(self, consequent, instance, substitutions):
        if is_aggregation_predicate(consequent):
            return substitutions

        return super().eliminate_already_computed(
            consequent, instance, substitutions
        )



import operator
from typing import AbstractSet

import numpy as np
import pytest

from ..exceptions import NeuroLangException
from ..expressions import Constant, Symbol
from ..probabilistic.cplogic import testing
from ..relational_algebra import (
    ColumnStr,
    ColumnInt,
    NaturalJoin,
    Product,
    RenameColumn,
    RenameColumns,
    Selection,
    eq_,
    str2columnstr_constant,
)
from ..relational_algebra_provenance import (
    ConcatenateConstantColumn,
    ExtendedProjection,
    ExtendedProjectionListMember,
    NaturalJoinInverse,
    Projection,
    ProvenanceAlgebraSet,
    RelationalAlgebraProvenanceCountingSolver,
    Union,
)
from ..utils import NamedRelationalAlgebraFrozenSet

C_ = Constant
S_ = Symbol

R1 = NamedRelationalAlgebraFrozenSet(
    columns=("col1", "col2", "__provenance__"),
    iterable=[(i, i * 2, i) for i in range(10)],
)
provenance_set_r1 = ProvenanceAlgebraSet(R1, ColumnStr("__provenance__"))


def test_selection():
    s = Selection(provenance_set_r1, eq_(C_(ColumnStr("col1")), C_(4)))
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s).value

    assert sol == R1.selection({"col1": 4})
    assert "__provenance__" in sol.columns


def test_selection_columns():
    s = Selection(
        provenance_set_r1, eq_(C_(ColumnStr("col1")), C_(ColumnStr("col2")))
    )
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s).value

    assert sol == R1.selection_columns({"col1": "col2"})
    assert sol == R1.selection({"col1": 0})
    assert "__provenance__" in sol.columns


def test_valid_rename():
    s = RenameColumn(
        provenance_set_r1, C_(ColumnStr("col1")), C_(ColumnStr("renamed"))
    )
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s).value

    assert sol == R1.rename_column("col1", "renamed")
    assert "renamed" in sol.columns
    assert "col1" not in sol.columns
    assert "__provenance__" in sol.columns
    assert R1.projection("__provenance__") == sol.projection("__provenance__")


def test_provenance_rename():
    s = RenameColumn(
        provenance_set_r1,
        C_(ColumnStr("__provenance__")),
        C_(ColumnStr("renamed")),
    )

    sol = RelationalAlgebraProvenanceCountingSolver().walk(s)

    assert sol.provenance_column == "renamed"
    sol = sol.value
    assert sol == R1.rename_column("__provenance__", "renamed")
    assert "renamed" in sol.columns
    assert "__provenance__" not in sol.columns


def test_naturaljoin():
    RA1 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "__provenance__"),
        iterable=[(i * 2, i) for i in range(10)],
    )
    pset_r1 = ProvenanceAlgebraSet(RA1, ColumnStr("__provenance__"))

    RA2 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "colA", "__provenance__"),
        iterable=[(i % 5, i * 3, i) for i in range(20)],
    )
    pset_r2 = ProvenanceAlgebraSet(RA2, ColumnStr("__provenance__"))

    s = NaturalJoin(pset_r1, pset_r2)
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s)

    RA1_np = NamedRelationalAlgebraFrozenSet(
        columns=("col1",), iterable=[(i * 2) for i in range(10)]
    )

    RA2_np = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "colA"), iterable=[(i % 5, i * 3) for i in range(20)]
    )

    R1njR2 = RA1_np.naturaljoin(RA2_np)
    sol_np = sol.value.projection(*["col1", "colA"])
    assert sol_np == R1njR2

    R1 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "__provenance__1"),
        iterable=[(i * 2, i) for i in range(10)],
    )

    R2 = NamedRelationalAlgebraFrozenSet(
        columns=("colA", "__provenance__2"),
        iterable=[(i * 3, i) for i in range(20)],
    )

    R1cpR2 = R1.cross_product(R2)
    RnjR = R1cpR2.naturaljoin(R1njR2)

    res = ExtendedProjection(
        ProvenanceAlgebraSet(RnjR, ColumnStr("__provenance__1")),
        tuple(
            [
                ExtendedProjectionListMember(
                    fun_exp=Constant(ColumnStr("__provenance__1"))
                    * Constant(ColumnStr("__provenance__2")),
                    dst_column=Constant(ColumnStr("__provenance__")),
                )
            ]
        ),
    )

    res = RelationalAlgebraProvenanceCountingSolver().walk(res)

    prov_sol = sol.value.projection("__provenance__")
    prov_res = res.value.projection("__provenance__")
    assert np.all(prov_sol == prov_res)


def test_naturaljoin_provenance_name():
    RA1 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "__provenance__1"),
        iterable=[(i * 2, i) for i in range(10)],
    )
    pset_r1 = ProvenanceAlgebraSet(RA1, ColumnStr("__provenance__1"))

    RA2 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "colA", "__provenance__2"),
        iterable=[(i % 5, i * 3, i) for i in range(20)],
    )
    pset_r2 = ProvenanceAlgebraSet(RA2, ColumnStr("__provenance__2"))

    s = NaturalJoin(pset_r1, pset_r2)
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s)

    assert sol.provenance_column == "__provenance__1"
    assert "__provenance__1" in sol.value.columns
    assert "__provenance__2" not in sol.value.columns


def test_product():
    RA1 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "__provenance__"),
        iterable=[(i * 2, i) for i in range(10)],
    )
    pset_r1 = ProvenanceAlgebraSet(RA1, ColumnStr("__provenance__"))

    RA2 = NamedRelationalAlgebraFrozenSet(
        columns=("colA", "__provenance__"),
        iterable=[(i * 3, i) for i in range(20)],
    )
    pset_r2 = ProvenanceAlgebraSet(RA2, ColumnStr("__provenance__"))

    s = Product((pset_r1, pset_r2))
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s).value

    RA1_np = NamedRelationalAlgebraFrozenSet(
        columns=("col1",), iterable=[(i * 2) for i in range(10)]
    )

    RA2_np = NamedRelationalAlgebraFrozenSet(
        columns=("colA",), iterable=[(i * 3) for i in range(20)]
    )

    R1njR2 = RA1_np.cross_product(RA2_np)
    sol_np = sol.projection(*["col1", "colA"])
    assert sol_np == R1njR2

    R1 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "__provenance__1"),
        iterable=[(i * 2, i) for i in range(10)],
    )

    R2 = NamedRelationalAlgebraFrozenSet(
        columns=("colA", "__provenance__2"),
        iterable=[(i * 3, i) for i in range(20)],
    )

    R1cpR2 = R1.cross_product(R2)
    RnjR = R1cpR2.naturaljoin(R1njR2)
    res = ExtendedProjection(
        ProvenanceAlgebraSet(RnjR, ColumnStr("__provenance__1")),
        tuple(
            [
                ExtendedProjectionListMember(
                    fun_exp=Constant(ColumnStr("__provenance__1"))
                    * Constant(ColumnStr("__provenance__2")),
                    dst_column=Constant(ColumnStr("__provenance__")),
                )
            ]
        ),
    )

    res = RelationalAlgebraProvenanceCountingSolver().walk(res)

    prov_sol = sol.projection("__provenance__")
    prov_res = res.value.projection("__provenance__")
    assert np.all(prov_sol == prov_res)


def test_product_provenance_name():
    RA1 = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "__provenance__1"),
        iterable=[(i * 2, i) for i in range(10)],
    )
    pset_r1 = ProvenanceAlgebraSet(RA1, ColumnStr("__provenance__1"))

    RA2 = NamedRelationalAlgebraFrozenSet(
        columns=("col2", "colA", "__provenance__2"),
        iterable=[(i % 5, i * 3, i) for i in range(20)],
    )
    pset_r2 = ProvenanceAlgebraSet(RA2, ColumnStr("__provenance__2"))

    s = Product((pset_r1, pset_r2))
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s)

    assert sol.provenance_column == "__provenance__1"
    assert "__provenance__1" in sol.value.columns
    assert "__provenance__2" not in sol.value.columns


def test_union():
    relation1 = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[
                ("a", "b", 1),
                ("b", "a", 2),
                ("c", "a", 2),
                ("c", "a", 1),
                ("b", "a", 1),
            ],
            columns=["x", "y", "__provenance__"],
        ),
        ColumnStr("__provenance__"),
    )
    relation2 = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[
                ("a", "b", 1),
                ("b", "a", 2),
                ("d", "a", 2),
                ("c", "a", 1),
                ("b", "a", 1),
            ],
            columns=["x", "y", "__provenance__"],
        ),
        ColumnStr("__provenance__"),
    )

    expected = NamedRelationalAlgebraFrozenSet(
        iterable=[("a", "b", 2), ("b", "a", 6), ("c", "a", 4), ("d", "a", 2)],
        columns=["x", "y", "__provenance__"],
    )

    s = Union(relation1, relation2)
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s).value

    assert sol == expected


def test_union_different_prov_col_names():
    r1 = testing.make_prov_set([(0.1, "a"), (0.2, "b")], ("_p1_", "x"))
    r2 = testing.make_prov_set([(0.5, "a"), (0.9, "c")], ("_p2_", "x"))
    expected = testing.make_prov_set(
        [(0.6, "a"), (0.2, "b"), (0.9, "c")], ("_whatever_", "x"),
    )
    operation = Union(r1, r2)
    solver = RelationalAlgebraProvenanceCountingSolver()
    result = solver.walk(operation)
    assert testing.eq_prov_relations(result, expected)


def test_union_with_empty_set():
    r = testing.make_prov_set([(0.1, "a"), (0.2, "b")], ("_p_", "x"))
    empty = testing.make_prov_set([], ("_p_", "x"))
    operation = Union(r, empty)
    solver = RelationalAlgebraProvenanceCountingSolver()
    result = solver.walk(operation)
    assert testing.eq_prov_relations(result, r)


def test_projection():
    relation = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[
                ("a", "b", 1),
                ("b", "a", 2),
                ("c", "a", 2),
                ("c", "a", 1),
                ("b", "a", 1),
            ],
            columns=["x", "y", "__provenance__"],
        ),
        ColumnStr("__provenance__"),
    )
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[("a", "b", 1), ("b", "a", 3), ("c", "a", 3)],
            columns=["x", "y", "__provenance__"],
        ),
        ColumnStr("__provenance__"),
    )
    sum_agg_op = Projection(
        relation, tuple([Constant(ColumnStr("x")), Constant(ColumnStr("y"))]),
    )
    solver = RelationalAlgebraProvenanceCountingSolver()
    result = solver.walk(sum_agg_op)

    assert result == expected


def test_concatenate_constant():
    s = ConcatenateConstantColumn(
        provenance_set_r1, C_(ColumnStr("new_col")), C_(9)
    )
    sol = RelationalAlgebraProvenanceCountingSolver().walk(s).value

    expected = NamedRelationalAlgebraFrozenSet(
        columns=("col1", "col2", "__provenance__", "new_col"),
        iterable=[(i, i * 2, i, 9) for i in range(10)],
    )

    assert expected == sol


def test_extended_projection():
    relation = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[(5, 1, 1), (6, 2, 2), (7, 3, 2), (1, 3, 1), (2, 1, 1)],
            columns=["x", "y", "__provenance__"],
        ),
        ColumnStr("__provenance__"),
    )

    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[(6, 1), (8, 2), (10, 2), (4, 1), (3, 1),],
            columns=["sum", "__provenance__"],
        ),
        ColumnStr("__provenance__"),
    )

    res = ExtendedProjection(
        relation,
        tuple(
            [
                ExtendedProjectionListMember(
                    fun_exp=Constant(ColumnStr("x"))
                    + Constant(ColumnStr("y")),
                    dst_column=Constant(ColumnStr("sum")),
                )
            ]
        ),
    )

    res = RelationalAlgebraProvenanceCountingSolver().walk(res)
    assert res == expected


def test_provenance_projection():
    relation = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[
                (0.8, "a", 42),
                (0.7, "b", 84),
                (0.2, "a", 21),
                (0.1, "b", 128),
            ],
            columns=["myprov", "x", "y"],
        ),
        ColumnStr("myprov"),
    )
    projection = Projection(relation, (Constant(ColumnStr("x")),))
    solver = RelationalAlgebraProvenanceCountingSolver()
    result = solver.walk(projection)
    assert len(result.value) == 2
    for exp_prob, exp_x in [(1.0, "a"), (0.8, "b")]:
        for tupl in result.value:
            if tupl.x == exp_x:
                assert np.isclose(
                    exp_prob, getattr(tupl, result.provenance_column)
                )


def test_provenance_product_with_shared_non_prov_col_should_fail():
    r1 = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[
                (0.8, "a", 42),
                (0.7, "b", 84),
                (0.2, "a", 21),
                (0.1, "b", 128),
            ],
            columns=["myprov", "x", "y"],
        ),
        ColumnStr("myprov"),
    )
    r2 = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            iterable=[(0.2, "a", 42), (0.5, "b", 84)],
            columns=["myprov", "x", "z"],
        ),
        ColumnStr("myprov"),
    )
    product = Product((r1, r2))
    solver = RelationalAlgebraProvenanceCountingSolver()
    with pytest.raises(
        NeuroLangException, match="Shared columns found: 'x'",
    ):
        solver.walk(product)


def test_concatenate_column_to_ra_relation():
    relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(
            iterable=[("a", "b"), ("c", "d")], columns=["x", "y"],
        )
    )
    solver = RelationalAlgebraProvenanceCountingSolver()
    concat_op = ConcatenateConstantColumn(
        relation, Constant(ColumnStr("z")), Constant[int](3)
    )
    result = solver.walk(concat_op)
    expected = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(
            iterable=[("a", "b", 3), ("c", "d", 3)], columns=["x", "y", "z"],
        )
    )
    assert result == expected


def test_rename_columns():
    prov_relation = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y"),
            iterable=[(0.1, "a", 0), (1.0, "b", 44)],
        ),
        provenance_column=ColumnStr("_p_"),
    )
    rename_columns = RenameColumns(
        prov_relation,
        ((str2columnstr_constant("x"), str2columnstr_constant("z")),),
    )
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "z", "y"),
            iterable=[(0.1, "a", 0), (1.0, "b", 44)],
        ),
        provenance_column=ColumnStr("_p_"),
    )
    solver = RelationalAlgebraProvenanceCountingSolver()
    result = solver.walk(rename_columns)
    assert testing.eq_prov_relations(result, expected)


def test_njoin_inverse():
    r1 = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x"), iterable=[(1.0, "a"), (0.5, "b")],
        ),
        ColumnStr("_p_"),
    )
    r2 = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x"), iterable=[(0.5, "b")],
        ),
        ColumnStr("_p_"),
    )
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x"), iterable=[(1.0, "b")],
        ),
        ColumnStr("_p_"),
    )
    op = NaturalJoinInverse(r1, r2)
    solver = RelationalAlgebraProvenanceCountingSolver()
    result = solver.walk(op)
    assert testing.eq_prov_relations(result, expected)


def test_selection_between_columnints():
    r = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y"),
            iterable=[
                (0.5, "a", "b"),
                (0.7, "a", "a"),
                (0.2, "b", "b"),
            ],
        ),
        ColumnStr("_p_"),
    )
    col1 = Constant[ColumnInt](ColumnInt(0))
    col2 = Constant[ColumnInt](ColumnInt(1))
    op = Selection(r, Constant(operator.eq)(col1, col2))
    solver = RelationalAlgebraProvenanceCountingSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y"),
            iterable=[
                (0.7, "a", "a"),
                (0.2, "b", "b"),
            ],
        ),
        ColumnStr("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)



import pytest

from ...datalog import Fact
from ...exceptions import ForbiddenExpressionError
from ...expressions import Constant, Symbol
from ...logic import Implication, Union
from ...logic.unification import apply_substitution, most_general_unifier
from ..ppdl import PPDL, PPDLDeltaTerm, get_dterm

C_ = Constant
S_ = Symbol

x = S_("x")
y = S_("y")
z = S_("z")
P = S_("P")
Q = S_("Q")
R = S_("R")
W = S_("W")
K = S_("K")
Z = S_("Z")
p = S_("p")
q = S_("q")
a = C_("a")
b = C_("b")

bernoulli = S_("bernoulli")


def test_get_dterm():
    datom = P(x, y, PPDLDeltaTerm(S_("Hi"), tuple()))
    assert get_dterm(datom) == PPDLDeltaTerm(S_("Hi"), tuple())
    datom = P(x, y, PPDLDeltaTerm(S_("Hi"), (C_(2),)))
    assert get_dterm(datom) == PPDLDeltaTerm(S_("Hi"), (C_(2),))


def test_apply_substitution_to_delta_term():
    dterm = PPDLDeltaTerm(bernoulli, (p,))
    new_dterm = apply_substitution(dterm, {p: q})
    assert new_dterm == PPDLDeltaTerm(bernoulli, (q,))

    substitution = {S_("random_symbol"): S_("another_random_symbol")}
    new_dterm = apply_substitution(dterm, substitution)
    assert new_dterm == dterm


def test_apply_substitution_to_delta_atom():
    datom = P(x, PPDLDeltaTerm(bernoulli, (p,)))
    new_datom = apply_substitution(datom, {p: q})
    assert new_datom == P(x, PPDLDeltaTerm(bernoulli, (q,)))


def test_unification_of_delta_atom():
    a = P(x, PPDLDeltaTerm(bernoulli, (p,)))
    b = P(y, PPDLDeltaTerm(bernoulli, (q,)))
    mgu = most_general_unifier(a, b)
    assert mgu is not None
    unifier, _ = mgu
    assert unifier == {x: y, p: q}


def test_ppdl_program():
    tau_1 = Implication(P(x, PPDLDeltaTerm(bernoulli, (C_(0.5),))), Q(x))
    program = Union((tau_1,))
    ppdl = PPDL()
    ppdl.walk(program)
    edb = ppdl.extensional_database()
    idb = ppdl.intensional_database()
    assert tau_1 in idb[P].formulas

    with pytest.raises(ForbiddenExpressionError):
        tau_2 = Implication(
            P(
                x,
                PPDLDeltaTerm(bernoulli, tuple()),
                PPDLDeltaTerm(S_("Flap"), tuple()),
            ),
            Q(x),
        )
        ppdl = PPDL()
        ppdl.walk(Union((tau_2, Fact(Q(C_(2))))))
        edb = ppdl.extensional_database()
        idb = ppdl.intensional_database()
        assert tau_2 in idb[P].formulas
        assert Fact(Q(C_(2))) in edb[Q]


def test_burglar():
    City = S_("City")
    House = S_("House")
    Business = S_("Business")
    Unit = S_("Unit")
    Earthquake = S_("Earthquake")
    Burglary = S_("Burglary")
    Trig = S_("Trig")
    Alarm = S_("Alarm")
    x, h, b, c, r = S_("x"), S_("h"), S_("b"), S_("c"), S_("r")
    program = Union(
        (
            Implication(Unit(h, c), House(h, c)),
            Implication(Unit(b, c), Business(b, c)),
            Implication(
                Earthquake(c, PPDLDeltaTerm(bernoulli, (C_(0.01),))),
                City(c, r),
            ),
            Implication(
                Burglary(x, c, PPDLDeltaTerm(bernoulli, (r,))),
                Unit(x, c) & City(c, r),
            ),
            Implication(
                Trig(x, PPDLDeltaTerm(bernoulli, (C_(0.6),))),
                Unit(x, c) & Earthquake(c, C_(1)),
            ),
            Implication(
                Trig(x, PPDLDeltaTerm(bernoulli, (C_(0.9),))),
                Burglary(x, c, C_(1)),
            ),
            Implication(Alarm(x), Trig(x, C_(1))),
        )
    )


def test_pcs_example():
    Uniform = S_("uniform")
    Gender = S_("Gender")
    Subject = S_("Subject")
    pGender = S_("pGender")
    pHasLPC = S_("pHasLPC")
    pHasRPC = S_("pHasRPC")
    HasLPC = S_("HasLPC")
    HasRPC = S_("HasRPC")
    x = S_("x")
    p = S_("p")
    program = Union(
        (
            Implication(
                Gender(x, PPDLDeltaTerm(bernoulli, (p,))),
                Subject(x) & pGender(p),
            ),
            Implication(
                HasLPC(x, PPDLDeltaTerm(bernoulli, (p,))),
                Subject(x) & pHasLPC(p),
            ),
            Implication(
                HasRPC(x, PPDLDeltaTerm(bernoulli, (p,))),
                Subject(x) & pHasRPC(p),
            ),
            Implication(
                pGender(PPDLDeltaTerm(Uniform, (C_(0), C_(1)))), C_("True")
            ),
            Implication(
                pHasLPC(PPDLDeltaTerm(Uniform, (C_(0), C_(1)))), C_("True")
            ),
            Implication(
                pHasRPC(PPDLDeltaTerm(Uniform, (C_(0), C_(1)))), C_("True")
            ),
        )
    )



# coding: utf-8
r'''
Loading and Querying the Destrieux et al. Atlas' Left Hemisphere
========================================================================


Uploading the Destrieux regions NeuroLang and
executing a simple query.
'''

from matplotlib import pyplot as plt
import nibabel as nib
from nilearn import datasets, plotting
from neurolang.frontend import NeurolangDL


##################################################
# Initialise the NeuroLang probabilistic engine.

nl = NeurolangDL()


###############################################################################
# Load the Destrieux example from nilearn as a fact list


atlas_destrieux = datasets.fetch_atlas_destrieux_2009()
atlas_labels = {
    label: str(name.decode('utf8'))
    for label, name in atlas_destrieux['labels']
}


nl.add_atlas_set('destrieux_atlas', atlas_labels, nib.load(atlas_destrieux['maps']))

###############################################################################
# Add utility function


@nl.add_symbol
def startswith(prefix: str, s: str) -> bool:
    return s.startswith(prefix)

###############################################################################
# Query all left hemisphere regions superior to the temporal superior sulucs
# and anterior to the central sulcus


with nl.scope as e:
    e.on_left_hemisphere[e.x] = (
            e.destrieux_atlas(e.l, e.x) & e.startswith('L S', e.l)
    )

    result = nl.query((e.r,), e.on_left_hemisphere(e.r))


###############################################################################
# Visualise results

subplots = plt.subplots(
    nrows=len(result), ncols=1,
    figsize=(10, 5 * len(result))
)[1]

for (r, subplot) in zip(result, subplots):
    region = r[0]
    plotting.plot_roi(region.spatial_image(), axes=subplot)



from ....expressions import Symbol, Constant as _C
from ....expression_walker import PatternWalker, add_match
from ..chart_parser import ChartParser
from ..english_grammar import (
    EnglishGrammar,
    EnglishBaseLexicon,
    S,
    NP,
    PN,
    VP,
    V,
    DET,
    N,
    num,
    gen,
    PRO,
    case,
    VAR,
)
import pytest


_eg = EnglishGrammar
_cp = ChartParser(_eg, EnglishBaseLexicon())


def test_indefinite_noun_phrase():
    t = _cp.parse("every book has an ending")[0]
    assert t == S(num.singular)(
        NP(num.singular, gen.thing, case.nom)(
            DET(num.singular)(_C("every")),
            N(num.singular, gen.thing)(_C("book")),
        ),
        VP(num.singular)(
            V(num.singular)(_C("has")),
            NP(num.singular, gen.thing, case.notnom)(
                DET(num.singular)(_C("an")),
                N(num.singular, gen.thing)(_C("ending")),
            ),
        ),
    )


@pytest.mark.skip(reason="Not priority now")
def test_nested_variable_unification():
    t = _cp.parse("Jones and a man like the book")[0]
    g = t.args[0].functor.args[1]
    e = S(num.plural)(
        NP(num.plural, g, case.nom)(
            NP(num.singular, gen.male, case.nom)(
                PN(num.singular, gen.male)(_C("Jones"))
            ),
            _C("and"),
            NP(num.singular, gen.male, case.nom)(
                DET(num.singular)(_C("a")),
                N(num.singular, gen.male)(_C("man")),
            ),
        ),
        VP(num.plural)(
            V(num.plural)(_C("like")),
            NP(num.singular, gen.thing, case.notnom)(
                DET(num.singular)(_C("the")),
                N(num.singular, gen.thing)(_C("book")),
            ),
        ),
    )
    assert t == e


def test_pronouns():
    t = _cp.parse("she owns it")[0]
    e = S(num.singular)(
        NP(num.singular, gen.female, case.nom)(
            PRO(num.singular, gen.female, case.nom)(_C("she"))
        ),
        VP(num.singular)(
            V(num.singular)(_C("owns")),
            NP(num.singular, gen.thing, case.notnom)(
                PRO(num.singular, gen.thing, case.notnom)(_C("it")),
            ),
        ),
    )
    assert t == e


def test_pronouns_2():
    assert not _cp.recognize("it owns he")


def test_apposition():
    t = _cp.parse("the man X owns a book Y")[0]
    e = S(num.singular)(
        NP(num.singular, gen.male, case.nom)(
            NP(num.singular, gen.male, case.nom)(
                DET(num.singular)(_C("the")),
                N(num.singular, gen.male)(_C("man")),
            ),
            VAR()(_C("X")),
        ),
        VP(num.singular)(
            V(num.singular)(_C("owns")),
            NP(num.singular, gen.thing, case.notnom)(
                NP(num.singular, gen.thing, case.notnom)(
                    DET(num.singular)(_C("a")),
                    N(num.singular, gen.thing)(_C("book")),
                ),
                VAR()(_C("Y")),
            ),
        ),
    )
    assert t == e


def test_variable_reference():
    t = _cp.parse("X owns Y")[0]
    fresh1 = t.args[0].functor.args[1]
    fresh2 = t.args[1].args[1].functor.args[0]
    fresh3 = t.args[1].args[1].functor.args[1]
    e = S(num.singular)(
        NP(num.singular, fresh1, case.nom)(VAR()(_C("X"))),
        VP(num.singular)(
            V(num.singular)(_C("owns")),
            NP(fresh2, fresh3, case.notnom)(VAR()(_C("Y"))),
        ),
    )
    assert t == e


def test_implication_if():
    t = _cp.parse("X intersects Y if Y intersects X")[0]
    fresh = (
        t.functor.args[0],
        t.args[0].args[0].functor.args[1],
        t.args[0].args[1].args[1].functor.args[0],
        t.args[0].args[1].args[1].functor.args[1],
        t.args[2].args[0].functor.args[1],
        t.args[2].args[1].args[1].functor.args[0],
        t.args[2].args[1].args[1].functor.args[1],
    )
    e = S(fresh[0])(
        S(num.singular)(
            NP(num.singular, fresh[1], case.nom)(VAR()(_C("X"))),
            VP(num.singular)(
                V(num.singular)(_C("intersects")),
                NP(fresh[2], fresh[3], case.notnom)(VAR()(_C("Y"))),
            ),
        ),
        _C("if"),
        S(num.singular)(
            NP(num.singular, fresh[4], case.nom)(VAR()(_C("Y"))),
            VP(num.singular)(
                V(num.singular)(_C("intersects")),
                NP(fresh[5], fresh[6], case.notnom)(VAR()(_C("X"))),
            ),
        ),
    )
    assert t == e


def test_implication_if_then():
    t = _cp.parse("if Y intersects X then X intersects Y")[0]
    fresh = (
        t.functor.args[0],
        t.args[1].args[0].functor.args[1],
        t.args[1].args[1].args[1].functor.args[0],
        t.args[1].args[1].args[1].functor.args[1],
        t.args[3].args[0].functor.args[1],
        t.args[3].args[1].args[1].functor.args[0],
        t.args[3].args[1].args[1].functor.args[1],
    )
    e = S(fresh[0])(
        _C("if"),
        S(num.singular)(
            NP(num.singular, fresh[1], case.nom)(VAR()(_C("Y"))),
            VP(num.singular)(
                V(num.singular)(_C("intersects")),
                NP(fresh[2], fresh[3], case.notnom)(VAR()(_C("X"))),
            ),
        ),
        _C("then"),
        S(num.singular)(
            NP(num.singular, fresh[4], case.nom)(VAR()(_C("X"))),
            VP(num.singular)(
                V(num.singular)(_C("intersects")),
                NP(fresh[5], fresh[6], case.notnom)(VAR()(_C("Y"))),
            ),
        ),
    )
    assert t == e


def test_sentence_conjunction():
    _cp.parse("Jones owns Ulysses and Smith likes Ulysses")
    _cp.parse("Jones owns Ulysses, Smith owns Ulysses, and Smith owns Ulysses")
    _cp.parse(
        "Jones and Smith own Ulysses, "
        + "Jones likes Ulysses and Odyssey, "
        + "and Smith likes every book"
    )
    _cp.parse(
        """
        if a region X provides a function F and X intersects
        a region Y then Y affects F
        """
    )
    _cp.parse(
        """
        if a region provides a function and that region intersects
        a region Y then Y affects that function
        """
    )
    _cp.parse(
        """
        if a region provides a function, that region intersects
        a region Y, and Ulysses references Y then if Jones owns
        Ulysses then the region Y affects that function
        """
    )



from ....expressions import Symbol, Constant
from ....logic import (
    Implication,
    Conjunction,
    ExistentialPredicate,
    UniversalPredicate,
)
from ..drs_builder import DRSBuilder, DRS2FOL
from ..chart_parser import ChartParser
from ..english_grammar import EnglishGrammar, EnglishBaseLexicon
import pytest


_eg = EnglishGrammar
_cp = ChartParser(_eg, EnglishBaseLexicon())


def test_simple_expression():
    b = DRSBuilder(_eg)
    t = _cp.parse("Jones owns Ulysses")[0]
    drs = b.walk(t)

    assert len(drs.referents) == 0
    assert len(drs.expressions) == 1

    assert drs.expressions[0] == Symbol("owns")(
        Constant("Jones"), Constant("Ulysses")
    )


def test_indefinite_noun_phrase():
    b = DRSBuilder(_eg)
    t = _cp.parse("Jones owns a book")[0]
    drs = b.walk(t)

    assert len(drs.referents) == 1
    x = drs.referents[0]
    assert len(drs.expressions) == 2

    assert drs.expressions[0] == Symbol("owns")(Constant("Jones"), x)
    assert drs.expressions[1] == Symbol("book")(x)


def test_var_noun_phrases():
    b = DRSBuilder(_eg)
    t = _cp.parse("X intersects Y")[0]
    drs = b.walk(t)

    assert len(drs.referents) == 2
    assert len(drs.expressions) == 1
    assert drs.expressions[0] == Symbol("intersects")(Symbol("X"), Symbol("Y"))


def test_apposition_variable_introduction():
    b = DRSBuilder(_eg)
    t = _cp.parse("a region X intersects a region Y")[0]
    drs = b.walk(t)
    x = Symbol("X")
    y = Symbol("Y")
    assert len(drs.referents) == 2
    assert x == drs.referents[0]
    assert y == drs.referents[1]
    assert len(drs.expressions) == 3

    assert drs.expressions[0] == Symbol("intersects")(x, y)
    assert drs.expressions[1] == Symbol("region")(y)
    assert drs.expressions[2] == Symbol("region")(x)


def test_conditional():
    b = DRSBuilder(_eg)
    t = _cp.parse("if a region Y intersects a region X then X intersects Y")[0]
    drs = b.walk(t)
    x = Symbol("X")
    y = Symbol("Y")
    assert len(drs.expressions) == 1
    assert isinstance(drs.expressions[0], Implication)
    ant = drs.expressions[0].antecedent
    con = drs.expressions[0].consequent
    assert len(ant.referents) == 2
    assert set(ant.referents) == {x, y}

    assert Symbol("intersects")(y, x) in set(ant.expressions)
    assert Symbol("region")(x) in set(ant.expressions)
    assert Symbol("region")(y) in set(ant.expressions)

    assert len(con.referents) == 0
    assert Symbol("intersects")(x, y) in set(con.expressions)


def test_translation_1():
    b = DRSBuilder(_eg)
    t = _cp.parse("X intersects Y")[0]
    drs = b.walk(t)
    exp = DRS2FOL().walk(drs)
    x = Symbol("X")
    y = Symbol("Y")

    assert exp == ExistentialPredicate(
        y, ExistentialPredicate(x, Symbol("intersects")(x, y))
    )


def test_translation_2():
    b = DRSBuilder(_eg)
    t = _cp.parse("if a region Y intersects a region X then X intersects Y")[0]
    drs = b.walk(t)
    exp = DRS2FOL().walk(drs)
    x = Symbol("X")
    y = Symbol("Y")

    assert exp == UniversalPredicate(
        x,
        UniversalPredicate(
            y,
            Implication(
                Symbol("intersects")(x, y),
                Conjunction(
                    (
                        Symbol("intersects")(y, x),
                        Symbol("region")(x),
                        Symbol("region")(y),
                    )
                ),
            ),
        ),
    )


def test_same_implication():
    # Check that the next translate to the same:
    # - Jones likes a book X if X references Odyssey.
    # - if a book X references Odyssey then Jones likes X.
    pytest.skip("Incomplete")


def test_quoted_predicate():
    b = DRSBuilder(_eg)
    t = _cp.parse("if `intersects(Y, X)` then X intersects Y")[0]
    drs = b.walk(t)

    x = Symbol("X")
    y = Symbol("Y")

    assert len(drs.expressions) == 1
    assert isinstance(drs.expressions[0], Implication)
    ant = drs.expressions[0].antecedent
    con = drs.expressions[0].consequent
    assert len(ant.referents) == 2
    assert set(ant.referents) == {x, y}
    assert Symbol("intersects")(y, x) in set(ant.expressions)
    assert len(con.referents) == 0
    assert Symbol("intersects")(x, y) in set(con.expressions)


def test_conjunction_1():
    b = DRSBuilder(_eg)
    t = _cp.parse("X owns Y and Y references Z")[0]
    drs = b.walk(t)
    exp = DRS2FOL().walk(drs)
    x = Symbol("X")
    y = Symbol("Y")
    z = Symbol("Z")

    assert exp == ExistentialPredicate(
        z,
        ExistentialPredicate(
            y,
            ExistentialPredicate(
                x,
                Conjunction(
                    (Symbol("owns")(x, y), Symbol("references")(y, z),)
                ),
            ),
        ),
    )


def test_conjunction_2():
    b = DRSBuilder(_eg)
    t = _cp.parse(
        "a man X owns a book Y, Y references a book Z, and X likes Z"
    )[0]
    drs = b.walk(t)
    exp = DRS2FOL().walk(drs)
    x = Symbol("X")
    y = Symbol("Y")
    z = Symbol("Z")

    assert exp == ExistentialPredicate(
        z,
        ExistentialPredicate(
            y,
            ExistentialPredicate(
                x,
                Conjunction(
                    (
                        Symbol("owns")(x, y),
                        Symbol("book")(y),
                        Symbol("man")(x),
                        Symbol("references")(y, z),
                        Symbol("book")(z),
                        Symbol("likes")(x, z),
                    )
                ),
            ),
        ),
    )


def test_quoted_string_literal():
    b = DRSBuilder(_eg)
    t = _cp.parse('"Ulysses" references "Odyssey"')[0]
    drs = b.walk(t)
    o = Constant("Odyssey")
    u = Constant("Ulysses")

    assert len(drs.referents) == 0
    assert len(drs.expressions) == 1
    assert Symbol("references")(u, o) in set(drs.expressions)
